#!/usr/bin/env python
#
#  For more information, please see: http://software.sci.utah.edu
# 
# The MIT License
#
# Copyright (c) 2009 Scientific Computing and Imaging Institute,
# University of Utah.
#
#
# Permission is hereby granted, free of charge, to any person obtaining a 
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation 
# the rights to use, copy, modify, merge, publish, distribute, sublicense, 
# and/or sell copies of the Software, and to permit persons to whom the 
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included 
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS 
# OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL 
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
# DEALINGS IN THE SOFTWARE.
#

import os
import sys
from sys import argv
import getopt
import subprocess
from subprocess import Popen
import re
import math
from math import sqrt

print_only = False
use_shell = True
if sys.platform == "win32" :
    use_shell = False
scirun_processes = []

def pause_for_scirun_cmd(scirun_arg) :
    global scirun_processes

    sci_run_cmmd = "%sscirun %s" % (binary_path, scirun_arg)
    print sci_run_cmmd
    p = subprocess.Popen(sci_run_cmmd,shell=True,bufsize=0,stdin=subprocess.PIPE,stdout=subprocess.PIPE,stderr=subprocess.PIPE)
    scirun_processes.append(p)
    done = False
    while not done:
        input_str = raw_input("Continue? (y/n)")
        if (input_str == "y"):
            done = True
        elif (input_str == "n"):
            print "Please close all SCIRun instances to exit..."
            # N.B.:  Volume rendering over a forwarded X11 ssh session may cause SCIRun to crash
            #        which will hang the python process while waiting for SCIRun to terminate.
            for proc in scirun_processes:
                proc.wait()
            sys.exit(0)

def parse_nrrd_for_voxel_radius(fname):

    # run unu head on the input file and capture the result
    p = subprocess.Popen(["%sunu head %s" % (binary_path, fname)],shell=True,stdin=subprocess.PIPE,stdout=subprocess.PIPE,stderr=subprocess.PIPE)
    (out,err) = p.communicate() 

    spacing = ('1','1','1')

    # compose a regular expression for parsing out NRRD space directions
    tri = "\s*\(([\-\d\.]+),\s*([\-\d\.]+),\s*([\-\d\.]+)\)\s*"
    spc_re = re.compile("space directions:%s%s%s" % (tri, tri, tri))
    res = spc_re.search(out)
    if (res != None):
        digits = res.groups()
        x_spacing = float(digits[0])+float(digits[3])+float(digits[6])
        y_spacing = float(digits[1])+float(digits[4])+float(digits[7])
        z_spacing = float(digits[2])+float(digits[5])+float(digits[8])  
        spacing = (x_spacing,y_spacing,z_spacing)
    else:
        # compose a recular expression for parsing out NRRD spacing
        pat = re.compile(r'spacings:\s*([\-\d\.]+)\s+([\-\d\.]+)\s+([\-\d\.]+)')
        res = pat.search(out)
        if (res != None):
            spacing = res.groups()

    # NOTE: the next line should be removed when spacing/orientation is
    #  fixed in the sizing field and particle sytem code.  For now, it's
    #  just here as a safety precaution.  Secondary note: the above
    #  code should be enhanced so NRRD lines that start with a '#' are
    #  ignored.
    # spacing = ('1','1','1')

    radius = sqrt(pow(float(spacing[0]),2)+pow(float(spacing[1]),2)+pow(float(spacing[2]),2))

    medial_axis_point_scale = radius * 4.0/15.0
    os.environ['MEDIAL_AXIS_POINT_SCALE'] = '%f' % medial_axis_point_scale
    particles_vector_scale = radius * 17.0/15.0
    os.environ['PARTICLES_VECTOR_SCALE'] = '%f' % particles_vector_scale
    particles_point_scale = radius * 4.0/15.0
    os.environ['PARTICLES_POINT_SCALE'] = '%f' % particles_point_scale
    
#    print "fname=%s  spacing=%s %s %s\n"%(fname,spacing[0],spacing[1],spacing[2])
#    print "point scale: " + os.environ['MEDIAL_AXIS_POINT_SCALE']
#    print "vec scale: " + os.environ['PARTICLES_VECTOR_SCALE']
#    print "point scale2: " + os.environ['PARTICLES_POINT_SCALE']

#    print radius
#    sys.exit(0)


def do_system(cmmd) :
    if print_only :
        print cmmd
    else :
        ret = -1
        try:
            ret = subprocess.call(cmmd, shell=use_shell)
            if (ret < 0):
                sys.exit(ret)
        except OSError, e:
            print "Execution failed with error ", e
            sys.exit(ret)

def usage() :
    print "Usage: BuildMesh.py [-sFirstStage:LastStage,--stages=FirstStage:LastStage] [-h,--help] [-i,--interactive] [-p, --binary-path] model_config";
    print "-s/--stages= allow a subset of the stages to be run.  The stages are:\n"
    print "1\tMake solo NRRDs from a label NRRD."
    print "2\tExtract the material surfaces from each solo NRRD."
    print "3\tCompute the medial axis for each material surface."
    print "4\tCompute the sizing field (representing local feature size)\n\tfrom the medial axis."
    print "5\tGenerate the initial sampling of the material interfaces."
    print "6\tFrom the seeds, run a particle system to optimize the\n\tsampling of the material interfaces."
    print "7\tGenerate a surface mesh for each of the materials, and\n\tfor all materials combined."
    print "8\tGenerate a volume mesh for each of the materials, and for\n\tall materials combined.\n"
    print "So, for example"
    print "\tBuildMesh.py -s1:3 my_model_config.py"
    print "would run stages 1 through 3 on the model specified by my_model_config.py"
    print "-i/--interactive launches a SCIRun network at the end of each relevant"
    print "pipeline stage to help the user visualize the results."
    print "-p/--binary-path Path to executable programs used in the BioMesh3d pipeline."
    print "If this option is not used, this path must be set in the system PATH."
    print "-h/--help prints this screen."


if __name__ == '__main__' :
    args = argv
# Check that we got the right arguments......
    try:
        opts, args = getopt.getopt(args[1:],"ihds:p:",["help","interactive","display-only","stages=","binary-path="])
    except getopt.GetoptError, err:
        # print help information and exit:
        print str(err) # will print something like "option -a not recognized"
        usage()
        sys.exit(-1)

    start_stage = 1
    stop_stage = 8
    is_interactive = False
    display_only = False

    scripts_dir = os.path.abspath(os.path.dirname(argv[0]))
    main_bin_dir, dummy = os.path.split(scripts_dir)     
    binary_path=""
    
    if sys.platform == "win32" :
        print "running on Windows"
        # are we in an installed directory?
        binary_path=main_bin_dir
        test_binary_path = os.path.join(main_bin_dir, "unu.exe")
        if not(os.path.exists(test_binary_path)) :
            # are we in a development build directory?
            binary_path = os.path.join(main_bin_dir, "release")
            if not(os.path.exists(binary_path)) :
                binary_path = os.path.join(main_bin_dir, "debug")
                if not(os.path.exists(binary_path)) :
                    binary_path=""
    else :
      print "running on Linux/OSX"
      binary_path = main_bin_dir
      if binary_path[len(binary_path) - 1] != os.sep:
        binary_path += os.sep   
    
    print "default binary_dir = %s " % binary_path

    for o,a in opts:
        if o in ["-s","--stages"]:
            start_stage_str,stop_stage_str = a.split(":")
            start_stage = int(start_stage_str)
            stop_stage = int(stop_stage_str)
        if o in ["-i","--interactive"]:
            is_interactive = True
        if o in ["-h","--help"]:
            usage()
            sys.exit(0)
        if o in ["-d","--display-only"]:
            display_only = True
        if o in ["-p","--binary-path"]:
            binary_path  = a

    if scripts_dir[len(scripts_dir) - 1] != os.sep:
        scripts_dir += os.sep
    
    model_config = ""
    if len(argv) < 2:
        print "Model configuration file is needed to run meshing pipeline."
        getUserInput = True
        while getUserInput:
            model_config = raw_input("Which model configuration file to use?  ")
            if (os.path.exists(model_config)):
                getUserInput = False
            else:
                print "Model configuration file ", model_config, " does not exist."
                input_str = raw_input("Try again? (y/n) ")
                if (input_str == "n"):
                    print "Exiting..."
                    sys.exit(0)
    else:
        model_config = args[0]
        if not(os.path.exists(model_config)):
            print "Model configuration file %s does not exist." % args[0]
            sys.exit(-1)
        
    model_config = os.path.abspath(model_config)
    
    execfile(model_config)
    model_path, dummy = os.path.split(model_config)
    
    if not(os.path.isabs(model_input_file)) :
      model_input_file = os.path.normpath(os.path.join(model_path, model_input_file))

    if not(os.path.isabs(model_output_path)) :
      model_output_path = os.path.normpath(os.path.join(model_path, model_output_path))
   

    print "*******************************************"
    print "model_config = %s" % model_config
    print "model_input_file = %s" % model_input_file
    print "model_output_path = %s" % model_output_path
    print "*******************************************"

    python_exec = sys.executable
    if len(python_exec) == 0:
        # try to supply a reasonable default
        python_exec = "python"

    if not os.path.exists(model_input_file) :
      print "The model_input_file %s does not exist" % model_input_file
      sys.exit(-1)

    if not os.path.exists(model_output_path) :
        os.makedirs(model_output_path)

# Done checking arguments

    print "Running stages %s through %s" % (start_stage,stop_stage)

    parse_nrrd_for_voxel_radius(model_input_file)

    
    print ""
    print "***************"
    print "Stage 1 -- Make Solo Nrrds"
    print "***************"
    current_stage = 1
    if ((start_stage <= current_stage) and (stop_stage >= current_stage)):
        solo_cmmd = r'"%s" "%sMakeSoloNrrd.py" "%s" "%s"' % (python_exec, scripts_dir, model_config, binary_path)
        if not display_only:
            do_system(solo_cmmd)
    
    print ""
    print "***************"
    print "Stage 2 -- Extract Material Surfaces"
    print "***************"
    current_stage += 1
    if (start_stage <= current_stage and stop_stage >= current_stage):
        bdry_cmmd = r'"%s" "%sComputeMaterialBoundary.py" "%s" "%s"' % (python_exec, scripts_dir, model_config, binary_path)
        if not display_only:
            do_system(bdry_cmmd)

        if (is_interactive):
            os.environ['COMPUTE_MATERIAL_BOUNDARY_OUTPUT'] = '%s/isosurface-all.ts.fld' % model_output_path
            print scripts_dir
            pause_for_scirun_cmd("-e %sMaterialBoundaries.srn" % scripts_dir)
    
    print ""
    print "***************"
    print "Stage 3 -- Compute Medial Axis"
    print "***************"
    current_stage += 1
    if (start_stage <= current_stage and stop_stage >= current_stage):
        ma_cmmd = r'"%s" "%sComputeMaterialMedialAxis.py" "%s" "%s"' % (python_exec, scripts_dir, model_config, binary_path)
        if not display_only:
            do_system(ma_cmmd)
        if (is_interactive):
            os.environ['COMPUTE_MATERIAL_BOUNDARY_OUTPUT'] = '%s/isosurface-all.ts.fld' % model_output_path
            os.environ['COMPUTE_MATERIAL_MEDIAL_AXIS_OUTPUT'] = '%s/ma-all.pc.fld' % model_output_path
            pause_for_scirun_cmd("-e %sMedialAxisAndBoundaries.srn" % scripts_dir)

    print ""
    print "***************"
    print "Stage 4 -- Compute Sizing Field"
    print "***************"
    current_stage += 1
    if (start_stage <= current_stage and stop_stage >= current_stage):
        sf_cmmd = r'"%s" "%sComputeSizingField.py" "%s" "%s"' % (python_exec, scripts_dir, model_config, binary_path)
        if not display_only:
            do_system(sf_cmmd)
        if (is_interactive):
            initial_mat_idx = min(1,len(mat_names)-1)
            initial_mat = mat_names[initial_mat_idx]
            os.environ['COMPUTE_SIZING_FIELD_OUTPUT'] = '%s/%s_sf.nrrd' % (model_output_path,initial_mat)

            os.environ['COMPUTE_MATERIAL_BOUNDARY_OUTPUT_SOLO'] = "%s/%s_isosurface.ts.fld" % (model_output_path,initial_mat)
            os.environ['COMPUTE_MATERIAL_MEDIAL_AXIS_OUTPUT_SOLO'] = "%s/%s_ma.pc.fld" % (model_output_path,initial_mat)

            pause_for_scirun_cmd("-e %sSizingFieldMedialAxisAndBoundaries.srn" % scripts_dir)

    print ""
    print "***************"
    print "Stage 5 -- Generate Initial Material Interface Sampling"
    print "***************"
    current_stage += 1
    if (start_stage <= current_stage and stop_stage >= current_stage):
        seeds_cmmd = r'"%s" "%sGenerateSeeds.py" "%s" "%s"' % (python_exec, scripts_dir, model_config, binary_path)
        if not display_only:
            do_system(seeds_cmmd)

    print ""
    print "***************"
    print "Stage 6 -- Run Particle System"
    print "***************"
    current_stage += 1
    if (start_stage <= current_stage and stop_stage >= current_stage):
        ps_cmmd = r'"%s" "%sDistributeParticles.py" "%s" "%s"' % (python_exec, scripts_dir, model_config, binary_path)
        if not display_only:
            do_system(ps_cmmd)
        if (is_interactive):
            os.environ['DISTRIBUTE_PARTICLES_PC_OUTPUT'] = '%s/junctions/particle-all.pc.fld' % model_output_path
            os.environ['DISTRIBUTE_PARTICLES_PCV_OUTPUT'] = '%s/junctions/particle-all.pcv.fld' % model_output_path
            pause_for_scirun_cmd("-e %sDistributedParticles.srn" % scripts_dir)
   
    print ""
    print "***************" 
    print "Stage 7 -- Generate Surface Mesh"
    print "***************"
    current_stage += 1
    if (start_stage <= current_stage and stop_stage >= current_stage):
        iface_cmmd = r'"%s" "%sBuildMaterialInterfaceMesh.py" "%s" "%s"' % (python_exec, scripts_dir, model_config, binary_path)
        if not display_only:
            do_system(iface_cmmd)

    print ""
    print "***************"
    print "Stage 8 -- Generate Volume Mesh"
    print "***************"
    current_stage += 1
    if (start_stage <= current_stage and stop_stage >= current_stage):
        vol_cmmd = r'"%s" "%sBuildVolumetricMesh.py" "%s" "%s"' % (python_exec, scripts_dir, model_config, binary_path)
        if not display_only:
            do_system(vol_cmmd)
        if (is_interactive):
            os.environ['BUILD_VOLUMETRIC_MESH_OUTPUT'] = '%s/junctions/tets-all_transformed.fld' % model_output_path
            pause_for_scirun_cmd("-e %sVolumetricMesh.srn" % scripts_dir)

    print ""
    print "**************************"
    print "Biomesh3D completed"
    print "**************************"
