/*
  For more information, please see: http://software.sci.utah.edu
  
  The MIT License

  Copyright (c) 2009 Scientific Computing and Imaging Institute,
  University of Utah.


  Permission is hereby granted, free of charge, to any person obtaining a 
  copy of this software and associated documentation files (the "Software"),
  to deal in the Software without restriction, including without limitation 
  the rights to use, copy, modify, merge, publish, distribute, sublicense, 
  and/or sell copies of the Software, and to permit persons to whom the 
  Software is furnished to do so, subject to the following conditions:

  The above copyright notice and this permission notice shall be included 
  in all copies or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS 
  OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL 
  THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
  DEALINGS IN THE SOFTWARE.
*/

//! Class definition of this one
#include <Core/Algorithms/Fields/MergeFields/JoinFields.h>
#include <Core/Algorithms/FieldArray/MergeMesh/ConvertCurveArrayToSurface.h>

//! Need to find out what type of field we are dealing with
#include <Core/Datatypes/FieldInformation.h>

//! SearchGridT
#include <Core/Containers/SearchGridT.h>

#include <float.h>

namespace SCIRunAlgo {


class CurveConnection {
  public:
    CurveConnection() : ref_cnt(0) {}
    
    // For Handles
    int ref_cnt;
  
    // Indices on one end and the one on the next slice
    std::vector<VMesh::Node::index_type> indices1_;
    std::vector<VMesh::Node::index_type> indices2_;

};

typedef Handle<CurveConnection>           CurveConnectionHandle;
typedef std::list<CurveConnectionHandle > CurveConnections;


class CurveSegments {
  public:
    // Constructor for structure
    CurveSegments(VMesh* mesh);

    // Destructor
    ~CurveSegments() {}

    // Find out which segments we need to connect
    // This takes two layers of segments and links them together
    // and returns a list of how to connect segments
    bool find_connections(CurveSegments& above, CurveConnections& connections, double radius);
    
  private:
    // Internal function that follows a path of a segmentation and puts
    // the nodes of it in a vector
    bool find_nodes(VMesh::Node::array_type& segments,
                    bool circular,
                    std::vector<VMesh::Node::index_type>& indices, 
                    int idx1, int idx2, int dir);


    // The underlying curve mesh
    VMesh*  vmesh_;
    
    // The direction at each node 
    std::vector<Vector>                               directions_;
    // The segment that each node is part of
    std::vector<int>                                  segment_;
    // The local index within the segment
    std::vector<int>                                  segment_index_;
    
    // List of segments
    std::list<VMesh::Node::array_type>                segments_;

    // Is the segment circular
    std::list<bool>                                   circular_;
};


CurveSegments::CurveSegments(VMesh* vmesh)
{
  // Build the lookup tables with the mesh
  // This should synchronize the tables with the curve mesh
  // Currently this is not used as CurveMesh does not have accelleration
  vmesh->synchronize(Mesh::NODE_NEIGHBORS_E|Mesh::FIND_CLOSEST_NODE_E);

  // Get the number of nodes in each segment
  VMesh::size_type num_nodes = vmesh->num_nodes();
  segment_.resize(num_nodes);

  // local storage while analysing data
  std::vector<Point>             points;
  Point                          point;
  VMesh::Node::array_type        neighbors;
  VMesh::Node::array_type        lneighbors;
  std::vector<VMesh::index_type> order[2];

  VMesh::index_type assign_count = 0;
  VMesh::index_type start_index = 0;
  VMesh::index_type cur_index = 0;
  VMesh::index_type prev_index = 0;

  // Loop through all nodes and try find how they are connected to other
  // neighboring nodes and try to identify segments

  // Counter for counting segments
  int seg = 0;

  while (assign_count < num_nodes)
  {
    // order is used to track the nodes along the two branches getting out
    // of the starting point
    order[0].clear();
    order[1].clear();
    
    // test whether the segment is a loop
    bool circular = false;
    int  circ_cnt = 0;
    
    // Find a starting point.
    // Every node unassigned to a segment is a good starting point
    for (VMesh::index_type k=0;k<num_nodes;k++)
    {
      // If segment_[k] == -1, the node has not yet been assigned
      // Use it as a starting point
      if (segment_[k] == -1) { start_index = k; break; }
    }
    
    // Assign start to new index type
    segment_[start_index] = seg; 
    
    // Update assignment counter
    assign_count++;
      
    order[0].push_back(start_index);
    
    // Get the neighbors of the starting point
    vmesh->get_neighbors(neighbors,VMesh::Node::index_type(start_index));
          
    // which neighbor to follow  
    size_t test_n = 0;
      
    for (int i=0; i<2; i++)
    {
      if (neighbors.size() > test_n)
      {
        prev_index = start_index;
        
        // check whether the direction we are looking in has not been assigned
        // if so increase counter
        while ((test_n < neighbors.size()) && ((segment_[neighbors[test_n]]) >= 0) 
                && segment_[neighbors[test_n]] != seg) test_n++;
      
        // if we found a valid search direction
        if (test_n < neighbors.size())
        {
          // If neighbor is already assigned this to this segment
          if (segment_[neighbors[test_n]] == seg) 
          {
            circ_cnt++;
            // if two neigbors have the same segment id, it must be circular
            // Note the first one is most likely the path where we came from 
            if (circ_cnt == 2) circular = true;
          }
          else
          {
            // get the neihbor node whose track we are following
            cur_index = neighbors[test_n];
            // Set neigbbor to same segment
            segment_[cur_index] = seg; 
            // increase node count, of nodes already visited
            assign_count++;

            // At node to list
            order[i].push_back(cur_index);
            
            // Find neighbor again
            vmesh->get_neighbors(lneighbors,VMesh::Node::index_type(cur_index));

            // Trace the route of the segments
            while ((assign_count < num_nodes)&&(lneighbors.size() > 1))
            {
              if (lneighbors[0] == prev_index)
              {
                prev_index = cur_index; cur_index = lneighbors[1];
              }
              else
              {
                prev_index = cur_index; cur_index = lneighbors[0];
              }
            
              if (segment_[cur_index] < 0)
              {
                segment_[cur_index] = seg; 
                assign_count++;
                
                order[i].push_back(cur_index);
                
                // get local neighbors for next tracking step
                vmesh->get_neighbors(lneighbors,VMesh::Node::index_type(cur_index));
              }
              else
              {
                // end the segment scan
                lneighbors.clear();
              }
            }
          }
        }
      }
    }

    // Expand segment list by adding a new array type in the list
    segments_.resize(segments_.size()+1);
    (*(segments_.end())).resize(order[0].size()+order[1].size());

    // A short cut to the newly create segment path
    VMesh::Node::array_type& segment = *(segments_.end());

    // As we started somewhere in the middle and tracked both ends
    // reorder it such that we can track the full path
    
    // cnt keeps a count of what the local index in the loop is
    size_t cnt = 0;
    
    // start at the end of this branch
    for (int p=static_cast<int>(order[1].size()-1); p>=0;p--)
    {
      // get the node index and push it back on the list
      segment.push_back(order[1][p]);
      // Indicate per node in which segment it is located and which position
      // it has in the queue 
      segment_index_[order[1][p]] = cnt; cnt++;
    }
    // continue to follow this one to the end
    for (size_t p=0; p<order[0].size(); p++)
    {
      // get the node index and push it back on the list
      segment.push_back(order[0][p]);
      // Indicate per node in which segment it is located and which position
      // it has in the queue 
      segment_index_[order[0][p]] = cnt; cnt++;
    }

    //  Check whether it is larger than 1
    // Segments that are smaller are ignored and should not be added
    if (segment.size() > 1)
    {
      // Get the center locations of all the nodes involved
      std::vector<Point> points;
      vmesh->get_centers(points,segment);

      if (!circular)
      {
        // Get the connection of the first node, which has only one neighbor
        directions_[segment[0]] = (points[1]-points[0]).normalize();
      
        // For all the middle ones take the average of the two neighboring ones
        for (size_t k=1; k< segment.size()-1; k++)
        {
          Vector v1 = (points[k]-points[k-1]); v1.normalize();
          Vector v2 = (points[k+1]-points[k]); v2.normalize();
          directions_[segment[k]] += v1;
          directions_[segment[k]] += v2;
          directions_[segment[k]].normalize();
        }
        // The last one only has one neighbor again
        directions_[segment[segment.size()-1]] = (points[segment.size()-1]-points[segment.size()-2]).normalize();
      
      }
      else
      {
        // Get the connection of the first node, which loops back to the last one
        Vector v1 = (points[1]-points[0]); v1.normalize();
        Vector v2 = (points[0]-points[segment.size()-1]); v2.normalize();
        directions_[segment[0]] = v1+v2;
        directions_[segment[0]].normalize();
        
        // For all the middle ones take the average of the two neighboring ones
        for (size_t k=1; k< segment.size()-1; k++)
        {
          v1 = (points[k]-points[k-1]); v1.normalize();
          v2 = (points[k+1]-points[k]); v2.normalize();
          directions_[segment[k]] = v1+v2;
          directions_[segment[k]].normalize();
        }
      
        // The last is again a special case
        v1 = (points[segment.size()-1]-points[segment.size()-2]); v1.normalize();
        v2 = (points[0]-points[segment.size()-1]); v2.normalize();
        directions_[segment[segment.size()-1]] = v1 + v2;
        directions_[segment[segment.size()-1]].normalize();
      }
    }
    circular_.push_back(circular);

    // continue to next segment in the list
    seg++;
  }
}


bool
CurveSegments::
find_nodes(VMesh::Node::array_type& segments,
           bool circular,
           std::vector<VMesh::Node::index_type>& indices, 
           int idx1,
           int idx2,
           int dir) 
{
  if (dir > 0)
  {
    if (idx1 > idx2)
    {
      if(!circular) return (false);
      for (int i=idx1; i<static_cast<int>(segments.size()); i++)
      {
        indices.push_back(segments[i]);
      }
      for (int i=0; i<idx2;i++)
      {
        indices.push_back(segments[i]);
      }
      return (true);
    }
    else if (idx2 < idx1)
    {
      for (int i=idx1; i<idx2; i++)
      {
        indices.push_back(segments[i]);
      }
      return (true);
    }
    return (false);
  }
  else
  {
    if (idx1 > idx2)
    {
      for (int i=idx1; i>idx2; i--)
      {
        indices.push_back(segments[i]);
      }
      return (true);
    }
    else if (idx1 < idx2)
    {
      if(!circular) return (false);
      for (int i=idx1; i>=0;i--)
      {
        indices.push_back(segments[i]);
      }
      for (int i=static_cast<int>(segments.size())-1; i>idx2; i--)
      {
        indices.push_back(segments[i]);
      }
      return (true);
    }
    return (false);  
  }
}



bool
CurveSegments::
find_connections(CurveSegments& above, CurveConnections& connections, double radius)
{
  // Get the above mesh structure
  VMesh* above_mesh = above.vmesh_;
  
  // Find how many segments are contained in each CurveConnection structure
  size_t above_nsegments = above.segments_.size();
  
  const std::vector<Vector>& above_directions = above.directions_;
  const std::vector<int>&    above_segment = above.segment_;
  const std::vector<int>&    above_segment_index = above.segment_index_;
  
  // Get iterators of segment list
  std::list<VMesh::Node::array_type>::iterator it = segments_.begin();
  std::list<VMesh::Node::array_type>::iterator it_end = segments_.end();
  
  std::list<bool>::iterator cit = circular_.begin();
  std::list<bool>::iterator cit_end = circular_.end();
  
  std::vector<int>          buffer_indices;
  std::vector<double>       buffer_distances;
  std::vector<int>          buffer_directions;
  std::vector<VMesh::Node::index_type>   closest_nodes;
  std::vector<double>       node_distances;
  std::vector<bool>         buffer_segment_in_range;
  
  int seg_cnt = 0;
  
  int*    indices_ptr = 0;
  int*    directions_ptr = 0;
  
  // loop over each segment and try to find corresponding points
  while (it != it_end)
  {
    // is this a circular segment ?
    bool circular = (*cit);
    
    // Determine size of segments
    size_t num_nodes = (*it).size();

    // Table for looking up what the closest node is for each segment
    buffer_indices.resize(above_nsegments*num_nodes,-1);
    buffer_directions.resize(above_nsegments*num_nodes,0);
    buffer_segment_in_range.resize(above_nsegments,false);
    
    indices_ptr = &(buffer_indices[0]);
    directions_ptr = &(buffer_directions[0]);
    
    // Loop over elements within segment
    VMesh::Node::array_type& segments = (*it);
    VMesh::Node::array_type::iterator nit = segments.begin();
    VMesh::Node::array_type::iterator nit_end = segments.end();
    
    Point p; Vector dir;

    while (nit != nit_end)
    {
      VMesh::Node::index_type idx = (*nit);
      // get the location of the node, so we can find the ones close to it
      vmesh_->get_center(p,idx);
      dir = directions_[idx];
      
      // fill out the table with maximum distances the table has a distance
      // to each segment in the above plain to account for splitting
      // wave fronts
      buffer_distances.resize(above_nsegments,DBL_MAX);

      // Find all nodes within a certain radius
      above_mesh->find_closest_nodes(node_distances,closest_nodes,p,radius);
      
      // Fill out the table for each node in the segment
      for (size_t j=0; j<closest_nodes.size(); j++)
      {
        double d = Dot(dir,above_directions[closest_nodes[j]]);
        if (Abs(d) > 0.3)
        {
          // Which segment is the jth closest point part of
          int seg = above_segment[closest_nodes[j]];
          
          // if this one is closer than others of the same segment
          // then mark that index and that distance
          if (node_distances[j] < buffer_distances[seg])
          {
            buffer_segment_in_range[seg] = true;
            buffer_distances[seg] = node_distances[j];
            indices_ptr[seg] = above_segment_index[closest_nodes[j]];
            // figure out in which direction to loop
            // this is useful for tying everything together
            if (d>0) directions_ptr[seg] = 1; 
            else directions_ptr[seg] = -1;
          } 
        }
      }      
            
      // We do not care about distances after this one and we can overwrite
      // the old ones.
      indices_ptr += above_nsegments;
      directions_ptr += above_nsegments;
    }
    
    directions_ptr = &(buffer_directions[0]);
    indices_ptr = &(buffer_indices[0]);

    std::list<VMesh::Node::array_type>::iterator ait = above.segments_.begin();
    std::list<VMesh::Node::array_type>::iterator ait_end = above.segments_.end();
    
    std::list<bool>::iterator acit = above.circular_.begin();
    std::list<bool>::iterator acit_end = above.circular_.end();    

    // scan for each segment whether we need to connect to it
    int scnt = 0;
    
    while (ait != ait_end)
    {
      // if the current segment does not overlap with the above segment
      // do not do any analysis on it
      if (!(buffer_segment_in_range[scnt])) { scnt++; continue; }

      VMesh::Node::array_type& above_segments = (*ait);
      bool above_circular = (*acit);
      
      int start = 0;
      int end = num_nodes-1;
      
      if (circular)
      {
        // Make sure there are boundaries between connections with the same segment in opposite
        // directions
        if (directions_ptr[scnt]*directions_ptr[scnt+above_nsegments*(num_nodes-1)] == -1)
        {
          directions_ptr[scnt+above_nsegments*(num_nodes-1)] = 0;
        }
        for (size_t i=0; i<num_nodes-1; i++)
        {
          if(directions_ptr[scnt+above_nsegments*i]*directions_ptr[scnt+above_nsegments*(i+1)] == -1)
          {
            directions_ptr[scnt+above_nsegments*(i+1)] = 0;
          }
        }
  
        // In case we have a circular segment crossing the boundary
        if (directions_ptr[scnt]*directions_ptr[scnt+above_nsegments*(num_nodes-1)] > 0)
        {
          while (start < static_cast<int>(num_nodes-1) && directions_ptr[scnt+above_nsegments*start] != 0) start++;
          while (end > 1 && directions_ptr[scnt+above_nsegments*end] != 0) end--;

          CurveConnectionHandle c = new CurveConnection();
          
          if (end < start)
          {
            // one big circular section
            for (size_t i=0; i<num_nodes-1; i++)
            {
              size_t y = i+1; if (y == num_nodes) y = 0;
              VMesh::Node::index_type idx1 = segments[i];

              c->indices1_.push_back(idx1);
              
              find_nodes(above_segments,
                         above_circular,
                         c->indices2_, 
                         indices_ptr[scnt+above_nsegments*i], 
                         indices_ptr[scnt+above_nsegments*y],
                         directions_ptr[scnt]);
            }
          }
          else
          {
            for (size_t i=end+1; i<num_nodes-1; i++)
            {
              size_t y = i+1; if (y == num_nodes) y = 0;
              VMesh::Node::index_type idx1 = segments[i];
              c->indices1_.push_back(idx1);
              
              if (directions_ptr[scnt+above_nsegments*y])
              {
                above.find_nodes(above_segments,
                                 above_circular,
                                 c->indices2_, 
                                 indices_ptr[scnt+above_nsegments*i], 
                                 indices_ptr[scnt+above_nsegments*y],
                                 directions_ptr[scnt+above_nsegments*i]);
              }
              else
              {
                c->indices2_.push_back(indices_ptr[scnt+above_nsegments*i]);              
              }
            }            
            for (int i=0; i<(start-2); i++)
            {
              size_t y = i+1;
              VMesh::Node::index_type idx1 = segments[i];
              VMesh::Node::index_type idx2 = segments[y];
              c->indices1_.push_back(idx1);
              
              if (directions_ptr[scnt+above_nsegments*y])
              {
                above.find_nodes(above_segments,
                                 above_circular,
                                 c->indices2_, 
                                 indices_ptr[scnt+above_nsegments*i], 
                                 indices_ptr[scnt+above_nsegments*y],
                                 directions_ptr[scnt+above_nsegments*i]); 
              }
              else
              {
                c->indices2_.push_back(indices_ptr[scnt+above_nsegments*i]);
              }

            }            

            // If they only contain one node on either side we skip them
            if (c->indices2_.size() > 1 && c->indices1_.size() > 1)
              connections.push_back(c);
          }
        }
      }
      
      // both circular and none circular cases

      for (int i=start; i<(end+1); i++)
      {
        if (directions_ptr[scnt+above_nsegments*i])
        {
          
          CurveConnectionHandle c = new CurveConnection();            
          
          while (i<(end+1) && directions_ptr[scnt+above_nsegments*i])
          {
            size_t y = i+1; if (y == num_nodes) y = 0;
            VMesh::Node::index_type idx1 = segments[i];
            VMesh::Node::index_type idx2 = segments[y];
            c->indices1_.push_back(idx1);
            
            if (directions_ptr[scnt+above_nsegments*y])
            {
              above.find_nodes(above_segments,
                               above_circular,
                               c->indices2_, 
                               indices_ptr[scnt+above_nsegments*i], 
                               indices_ptr[scnt+above_nsegments*y],
                               directions_ptr[scnt+above_nsegments*i]);
            }
            else
            {
              c->indices2_.push_back(indices_ptr[scnt+above_nsegments*i]);
            }
            i++;
          }
          if (c->indices2_.size() > 1 && c->indices1_.size() > 1)
            connections.push_back(c);            
        }
      }            
      
      ait++;
      acit++;
      scnt++;
    }

    // process next segment
    seg_cnt++; ++it; ++cit;
  }
  
  return (true);
}



bool 
ConvertCurveArrayToSurfaceAlgo::run(FieldArrayHandle input, FieldHandle& output)
{
/*
  // We do not do progress reporting, hence the lower level algorithm will do this
  algo_start("MergeFieldArray",false);

  if (input.get_rep() == 0)
  {
    error("No input FieldArray");
    algo_end(); return (false);
  }

  double radius = get_scalar("radius");

  // Get the input array
  std::vector<FieldHandle>& inputs = input->array();

  std::vector<std::vector<int> > segment;
  std::vector<std::vector<VMesh::Node::index_type> > segment_order;
  std::vector<std::vector<size_t> > segment_offset;
  
  VMesh::Node::array_type neighbors;
  VMesh::Node::array_type lneighbors;
  std::vector<VMesh::index_type> order[2];
  
  segment.resize(inputs.size());
  segment_order.resize(inputs.size());
  segment_offset.resize(inputs.size());
  
  // Phase one - sort the segments such that we know how it connects
  
  for(size_t j=0; j< inputs.size(); j++)
  {
    VMesh::index_type order_count = 0;
    int seg = 0;
    VMesh* vmesh = inputs[j]->vmesh();
    vmesh->synchronize(Mesh::NODE_NEIGHBORS_E|Mesh::FIND_CLOSEST_NODE_E);

    VMesh::size_type num_nodes = vmesh->num_nodes();
    segment[j].resize(num_nodes);
    segment_order[j].reserve(num_nodes);
    segment_offset[j].clear();
    
    for (VMesh::size_type k=0; k<num_nodes;k++) segment[j][k] = -1;
    
    VMesh::index_type assign_count = 0;
    VMesh::index_type start_index = 0;
    VMesh::index_type cur_index = 0;
    VMesh::index_type prev_index = 0;

    while (assign_count < num_nodes)
    {
      order[0].clear();
      order[1].clear();
      bool circular = false;
      int  circ_cnt = 0;
      
      for (VMesh::index_type k=0;k<num_nodes;k++)
      {
        if (segment[j][k] == -1) { start_index = k; break; }
      }
      
      segment[j][start_index] = seg; assign_count++;
      order[0].push_back(start_index);
    
      vmesh->get_neighbors(neighbors,VMesh::Node::index_type(start_index));
      
      size_t test_n = 0;
      
      for (int i=0; i<2; i++)
      {
        if (neighbors.size() > test_n)
        {
          prev_index = start_index;
          
          while ((test_n < neighbors.size()) && ((segment[j][neighbors[test_n]]) >= 0) 
                  && segment[j][neighbors[test_n]] != seg) test_n++;
        
          if (test_n < neighbors.size())
          {
            if (segment[j][neighbors[test_n]] == seg) 
            {
              circ_cnt++; 
              if (circ_cnt == 2) circular = true;
            }
            else
            {
              cur_index = neighbors[test_n];
              segment[j][cur_index] = seg; assign_count++;
              order[i].push_back(cur_index);
              
              vmesh->get_neighbors(lneighbors,VMesh::Node::index_type(cur_index));
              while ((assign_count < num_nodes)&&(lneighbors.size() > 1))
              {
                if (lneighbors[0] == prev_index)
                {
                  prev_index = cur_index; cur_index = lneighbors[1];
                }
                else
                {
                  prev_index = cur_index; cur_index = lneighbors[0];
                }
              
                if (segment[j][cur_index] < 0)
                {
                  segment[j][cur_index] = seg; assign_count++;
                  order[i].push_back(cur_index);
                  vmesh->get_neighbors(lneighbors,VMesh::Node::index_type(cur_index));

                }
                else
                {
                  // end the segment scan
                  lneighbors.clear();
                }
              }
            }
          }
        }
      }
      seg++;
      segment_offset[j].push_back(segment_order[j].size());

      for (int p=static_cast<int>(order[1].size()-1); p>=0;p--)
      {
        segment_order[j].push_back(order[1][p]);
      }
      for (size_t p=0; p<order[0].size(); p++)
      {
        segment_order[j].push_back(order[0][p]);
      }
      if (circular)
      {
        if (order[1].size()) segment_order[j].push_back(order[1][order[1].size()-1]);
        else if (order[0].size()) segment_order[j].push_back(order[0][0]);
      }
    }
    segment_offset[j].push_back(segment_order[j].size());
  }






  std::vector<VMesh::Node::index_type> closest_nodes;
  std::vector<int> closest_segments;
  
  std::vector<VMesh::Node::index_type> split_nodes;
  std::vector<int> split_segments;
  std::vector<double> split_distance;
  
  double max_distance = radius;

  FieldInformation fi("TriSurfMesh",1,"int");
  output = CreateField(fi);

  VField* ofield = output->vfield();
  VMesh*  omesh  = output->vmesh();

  VMesh::size_type offset1 = 0;
  VMesh::size_type offset2 = 0;
  
  if (inputs.size())
  {
    VMesh* vmesh = inputs[0]->vmesh();
    VMesh::size_type num_nodes = vmesh->num_nodes();
    Point p;
    for (VMesh::Node::index_type idx=0; idx < num_nodes; idx++)
    {
      vmesh->get_center(p,idx);
      omesh->add_node(p);
    }
  }

  for(size_t j=1; j< inputs.size(); j++)
  { 
    std::cerr << "j="<<j<<"\n";
  
    const std::vector<VMesh::Node::index_type> seg_order = segment_order[j-1];
    const std::vector<size_t>& seg_offset = segment_offset[j-1];
    const std::vector<int>& seg = segment[j-1];
    const std::vector<VMesh::Node::index_type> seg_order2 = segment_order[j];
    const std::vector<size_t>& seg_offset2 = segment_offset[j];
    const std::vector<int>& seg2 = segment[j];

    std::vector<Point> centers1;
    std::vector<Point> centers2;

    VMesh* vmesh2 = inputs[j]->vmesh();
    VMesh* vmesh1 = inputs[j-1]->vmesh();
    
    vmesh1->get_all_node_centers(centers1);
    vmesh2->get_all_node_centers(centers2);
    
    VMesh::size_type num_nodes1 = vmesh1->num_nodes();
    VMesh::size_type num_nodes2 = vmesh2->num_nodes();
    
    closest_nodes.resize(num_nodes1,-1);
    closest_segments.resize(num_nodes1,-1);
    split_nodes.resize(num_nodes1,-1);
    split_segments.resize(num_nodes1,-1);
    split_distance.resize(num_nodes1,DBL_MAX);

    Point p;
    for (VMesh::Node::index_type idx=0; idx < num_nodes2; idx++)
    {
      omesh->add_node(centers2[idx]);
    }
    offset2 += num_nodes1;    
    
    // Find closest nodes starting from bottom
    Point p1,p2;
    for (VMesh::Node::index_type idx=0; idx<num_nodes1;idx++)
    {
      vmesh1->get_center(p1,idx);
std::cerr << "p1="<<p1<<"\n";      
      if(vmesh2->find_closest_node(p2,closest_nodes[idx],p1,max_distance))
      {
std::cerr << "closest node "<< idx << " " << closest_nodes[idx] << "\n";     
        closest_segments[idx] = seg2[closest_nodes[idx]]; 
      }
    }

    double dist;
    for (VMesh::Node::index_type idx=0; idx<num_nodes2;idx++)
    {
      vmesh2->get_center(p1,idx);
      VMesh::Node::index_type node_idx;
      if (vmesh1->find_closest_node(dist,p2,node_idx,p1,max_distance))
      {
        if (dist < split_distance[node_idx])
        {
          if (closest_segments[node_idx] != seg2[idx])
          {
            split_nodes[node_idx] = idx;
            split_segments[node_idx] = seg2[idx];
            split_distance[node_idx] = dist;
          }
        }
      }
    }
  
    VMesh::Node::array_type nodes(3);
    size_t num_segments = (seg_offset.size()-1);
std::cerr << "num_segments="<<num_segments<<"\n";
    
    for (size_t k=0; k<num_segments; k++)
    {
std::cerr << "segment "<<k<<"\n";
      size_t start = seg_offset[k];
      size_t end =   seg_offset[k+1];

std::cerr << "start="<<start<<" end="<<end<<"\n";
      if (seg_order[start] == seg_order[end-1])
      { // circular segment
std::cerr << "CIRCULAR\n";
      }
      else
      { // non circular segment
      
        // Meshing part
        int idx1_start = -1, idx2_start = -1, idx1_end = -1, idx2_end = -1, estart, eend;
        VMesh::Node::index_type nidx_start =-1, nidx_end = -1;
        int s = -1;
        
        size_t m= start;
        while (m < end)
        {
          s = closest_segments[seg_order[m]];
          if (s >= 0)
          {
            idx1_start = m;
            nidx_start = closest_nodes[seg_order[m]];
            while (m < end && (closest_segments[seg_order[m]] == s)) m++;
            idx1_end = (m-1);
            nidx_end = closest_nodes[seg_order[m-1]];
          
            estart = seg_offset2[s];
            eend = seg_offset2[s+1];
std::cerr << "estart="<<estart<<" eend="<<eend<<"\n";
            for (size_t i=estart;i<eend;i++)
            {
              if (seg_order2[i] == nidx_start) idx2_start = i;
              if (seg_order2[i] == nidx_end) idx2_end = i;
            }

std::cerr << "idx1_start="<< idx1_start << " idx1_end="<<idx1_end<<"\n";         
std::cerr << "idx2_start="<< idx2_start << " idx2_end="<<idx2_end<<"\n";         
            if (idx2_start < 0 || idx2_end < 0)
            {
std::cerr << "ERROR\n";            
              error("Internal error in algorithm");        
              algo_end(); return (false);
            }
            
            
            //Now build tris
            
            size_t dir = 1;
            if (idx2_start> idx2_end) dir = -1;
            
            size_t n1 = idx1_start;
            size_t n2 = idx2_start;
            
            while (!((n1 == idx1_end) && (n2 == idx2_end)))
            {
std::cerr << "n1="<<n1<<" n2="<<n2<<" idx1_end="<< idx1_end << " idx2_end="<< idx2_end<<" seg_order.size()="<<seg_order.size()<<" seg_order2.size()="<<seg_order2.size()<<"\n";
              if (n1 == idx1_end)
              {
                nodes[0] = static_cast<VMesh::Node::index_type>(seg_order[n1]+offset1);
                nodes[2] = static_cast<VMesh::Node::index_type>(seg_order2[n2]+offset2);
                nodes[1] = static_cast<VMesh::Node::index_type>(seg_order2[n2+dir]+offset2);
                n2+=dir;
                omesh->add_elem(nodes);
              }
              else if (n2 == idx2_end)
              {
                nodes[0] = static_cast<VMesh::Node::index_type>(seg_order[n1]+offset1);
                nodes[1] = static_cast<VMesh::Node::index_type>(seg_order[n1+1]+offset1);
                nodes[2] = static_cast<VMesh::Node::index_type>(seg_order2[n2]+offset2);
                n1++;
                omesh->add_elem(nodes);            
              }
              else
              {
                Vector v1 = centers1[seg_order[n1]] - centers2[seg_order2[n2+dir]];
                Vector v2 = centers1[seg_order[n1+1]] - centers2[seg_order[n2]];
              
                if (v1.length() > v2.length() )
                {
                  nodes[0] = static_cast<VMesh::Node::index_type>(seg_order[n1]+offset1);
                  nodes[1] = static_cast<VMesh::Node::index_type>(seg_order[n1+1]+offset1);
                  nodes[2] = static_cast<VMesh::Node::index_type>(seg_order2[n2]+offset2);
                  n1++;
                  omesh->add_elem(nodes);              
                }
                else
                {
                  nodes[0] = static_cast<VMesh::Node::index_type>(seg_order[n1]+offset1);
                  nodes[2] = static_cast<VMesh::Node::index_type>(seg_order2[n2]+offset2);
                  nodes[1] = static_cast<VMesh::Node::index_type>(seg_order2[n2+dir]+offset2);
                  n2+=dir;
                  omesh->add_elem(nodes);                            
                }
              }
            }
              
          }
          else
          {
            m++;
          }
        }
      }
    }
  
    offset1 += num_nodes2;      
  }

  algo_end(); return (true);
*/

  // function no longer implemented
  return false;
}


} // end namespace SCIRunAlgo
