//
//  For more information, please see: http://software.sci.utah.edu
//
//  The MIT License
//
//  Copyright (c) 2009 Scientific Computing and Imaging Institute,
//  University of Utah.
//
//
//  Permission is hereby granted, free of charge, to any person obtaining a
//  copy of this software and associated documentation files (the "Software"),
//  to deal in the Software without restriction, including without limitation
//  the rights to use, copy, modify, merge, publish, distribute, sublicense,
//  and/or sell copies of the Software, and to permit persons to whom the
//  Software is furnished to do so, subject to the following conditions:
//
//  The above copyright notice and this permission notice shall be included
//  in all copies or substantial portions of the Software.
//
//  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
//  OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
//  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
//  THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
//  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
//  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
//  DEALINGS IN THE SOFTWARE.
//
//    File   : ToolManager.h
//    Author : Martin Cole
//    Date   : Thu May 25 08:01:56 2006

#if !defined(ToolManager_h)
#define ToolManager_h

#include <Core/Events/BaseTool.h>
#include <string>
#include <stack>
#include <map>

#include <Core/Events/share.h>
namespace SCIRun {

class BaseEvent;

using namespace std;

//! The ToolManager holds a prioritized set of stacks of tools.  Each of the
//! top() tools on the stacks recieve the events in priority order.  Tools
//! deeper down the stack do not recieve events.

class SCISHARE ToolManager
{
public:

  ToolManager(string name);
  ~ToolManager();

  //! return the tool name at the specified priority
  string query_tool_id(unsigned int pri) {
    ps_map_t::const_iterator iter = stacks_.find(pri);
    if (iter == stacks_.end()) {
      return "";
    }
    const ts_stack_t &s = (*iter).second;
    if (s.empty()) {
      return "";
    }

    return s.top()->get_name();
  }

  //! add the tool t on the stack with the specified priority.
  //! the ToolManager takes ownership of the added, tool, calling delete
  void add_tool(tool_handle_t t, unsigned priority);

  // Set the tool at this priority rather than maintaining a stack.
  void set_tool(tool_handle_t t, unsigned priority);

  tool_handle_t get_tool(unsigned priority);

  //! remove the top tool on the stack with the specified priority.
  void rm_tool(unsigned priority);

  //! propagate the event to each top() tool on all the stacks in
  //! priority order, returning the possibly modified event.
  event_handle_t  propagate_event(event_handle_t event);

private:
  BaseTool::propagation_state_e send_pointer_event(tool_handle_t,
                                                   event_handle_t) const;
  BaseTool::propagation_state_e send_key_event(tool_handle_t,
                                               event_handle_t) const;
  BaseTool::propagation_state_e send_window_event(tool_handle_t,
                                                  event_handle_t) const;
  BaseTool::propagation_state_e send_tm_notify_event(tool_handle_t,
						     event_handle_t) const;
  BaseTool::propagation_state_e send_command_event(tool_handle_t,
						   event_handle_t) const;

  typedef stack<tool_handle_t>                            ts_stack_t;
  typedef map<unsigned, ts_stack_t, less<unsigned> >      ps_map_t;

  ps_map_t            stacks_;
  string              name_;
};

} // namespace SCIRun
#endif // ToolManager_h
