/*
   For more information, please see: http://software.sci.utah.edu

   The MIT License

   Copyright (c) 2009 Scientific Computing and Imaging Institute,
   University of Utah.

   
   Permission is hereby granted, free of charge, to any person obtaining a
   copy of this software and associated documentation files (the "Software"),
   to deal in the Software without restriction, including without limitation
   the rights to use, copy, modify, merge, publish, distribute, sublicense,
   and/or sell copies of the Software, and to permit persons to whom the
   Software is furnished to do so, subject to the following conditions:

   The above copyright notice and this permission notice shall be included
   in all copies or substantial portions of the Software.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
   OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
   THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
   LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
   FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
   DEALINGS IN THE SOFTWARE.
*/



/*
 *  GeomTetra.cc:  A tetrahedra object
 *
 *  Written by:
 *   Steven G. Parker & David Weinstein
 *   Department of Computer Science
 *   University of Utah
 *   April 1994
 *
 */

#include <Core/Util/Debug.h>

#include <Core/Geom/DrawInfoOpenGL.h>
#include <Core/Geom/GeomTetra.h>
#include <Core/Geom/GeomLine.h>

namespace SCIRun {

GeomTetra::GeomTetra(const Point& p1, const Point& p2,
		     const Point& p3, const Point& p4)
: GeomObj(), p1(p1), p2(p2), p3(p3), p4(p4)
{
  DEBUG_CONSTRUCTOR("GeomTetra")
}

GeomTetra::GeomTetra(const GeomTetra& copy)
: GeomObj(copy), p1(copy.p1), p2(copy.p2), p3(copy.p3), p4(copy.p4)
{
  DEBUG_CONSTRUCTOR("GeomTetra")
}

GeomTetra::~GeomTetra()
{
  DEBUG_DESTRUCTOR("GeomTetra")
}

GeomObj* GeomTetra::clone()
{
    return new GeomTetra(*this);
}

void GeomTetra::get_bounds(BBox& bb)
{
  bb.extend(p1);
  bb.extend(p2);
  bb.extend(p3);
  bb.extend(p4);
}

void
GeomTetra::draw(DrawInfoOpenGL* di, Material* matl, double)
{
  if (!pre_draw(di, matl, 1)) return;
  di->polycount_+=4;

  switch(di->get_drawtype())
  {
  case DrawInfoOpenGL::WireFrame:
    if (di->currently_lit_)
    {
      Vector n1(Plane(p1, p2, p3).normal());
      glBegin(GL_LINE_STRIP);
      glNormal3d(n1.x(),n1.y(),n1.z());
      glVertex3d(p1.x(), p1.y(), p1.z());
      glVertex3d(p2.x(), p2.y(), p2.z());
      glVertex3d(p3.x(), p3.y(), p3.z());
      glVertex3d(p1.x(), p1.y(), p1.z());
      glVertex3d(p4.x(), p4.y(), p4.z());
      glVertex3d(p2.x(), p2.y(), p2.z());
      glVertex3d(p3.x(), p3.y(), p3.z());
      glVertex3d(p4.x(), p4.y(), p4.z());
      glEnd();
    }
    else
    {
      glBegin(GL_LINE_STRIP);
      glVertex3d(p1.x(), p1.y(), p1.z());
      glVertex3d(p2.x(), p2.y(), p2.z());
      glVertex3d(p3.x(), p3.y(), p3.z());
      glVertex3d(p1.x(), p1.y(), p1.z());
      glVertex3d(p4.x(), p4.y(), p4.z());
      glVertex3d(p2.x(), p2.y(), p2.z());
      glVertex3d(p3.x(), p3.y(), p3.z());
      glVertex3d(p4.x(), p4.y(), p4.z());
      glEnd();
    }
    break;
  case DrawInfoOpenGL::Flat:
    // this should be made into a tri-strip, but I couldn;t remember how.

    /* this can be done as a triangle strip using 8 vertices, or
     * as a triangle fan with 5 and 1 single triangle (8 verts)
     * I am doing the fan now (ordering is wierd with a tri-strip), but
     * will switch to the tri-strip when I can test it, if it's faster
     */ 
  case DrawInfoOpenGL::Gouraud:
    // this should be made into a tri-strip, but I couldn;t remember how.

    /*
     * These are actualy just flat shaded, to get "gourad" shading
     * you could average the facet normals for all the faces touching
     * a given vertex.  I don't think there is a faster way to do this
     * using flat shading.
     */
    if (di->currently_lit_)
    {
      Vector n1(Plane(p1, p2, p3).normal());
      Vector n2(Plane(p1, p2, p4).normal());
      Vector n3(Plane(p4, p2, p3).normal());
      Vector n4(Plane(p1, p4, p3).normal());

      glBegin(GL_TRIANGLES);
      glNormal3d(n1.x(), n1.y(), n1.z());
      glVertex3d(p1.x(), p1.y(), p1.z());
      glVertex3d(p2.x(), p2.y(), p2.z());
      glVertex3d(p3.x(), p3.y(), p3.z());
        
      glNormal3d(n2.x(), n2.y(), n2.z());
      glVertex3d(p1.x(), p1.y(), p1.z());
      glVertex3d(p2.x(), p2.y(), p2.z());
      glVertex3d(p4.x(), p4.y(), p4.z());

      glNormal3d(n3.x(), n3.y(), n3.z());
      glVertex3d(p4.x(), p4.y(), p4.z());
      glVertex3d(p2.x(), p2.y(), p2.z());
      glVertex3d(p3.x(), p3.y(), p3.z());

      glNormal3d(n4.x(), n4.y(), n4.z());
      glVertex3d(p1.x(), p1.y(), p1.z());
      glVertex3d(p4.x(), p4.y(), p4.z());
      glVertex3d(p3.x(), p3.y(), p3.z());
      glEnd();
    }
    else
    {
      glBegin(GL_TRIANGLE_FAN);
      glVertex3d(p1.x(), p1.y(), p1.z());
      glVertex3d(p2.x(), p2.y(), p2.z());
      glVertex3d(p3.x(), p3.y(), p3.z());
      glVertex3d(p4.x(), p4.y(), p4.z());
      glVertex3d(p2.x(), p2.y(), p2.z());
      glEnd();
      glBegin(GL_TRIANGLES);
      glVertex3d(p4.x(), p4.y(), p4.z());
      glVertex3d(p2.x(), p2.y(), p2.z());
      glVertex3d(p3.x(), p3.y(), p3.z());
      glEnd();
    }
#if 0
    /*
     * This has inconsistant ordering.
     */
    glBegin(GL_TRIANGLES);
    glVertex3d(p1.x(), p1.y(), p1.z());
    glVertex3d(p2.x(), p2.y(), p2.z());
    glVertex3d(p3.x(), p3.y(), p3.z());
    glEnd();
    glBegin(GL_TRIANGLES);
    glVertex3d(p1.x(), p1.y(), p1.z());
    glVertex3d(p2.x(), p2.y(), p2.z());
    glVertex3d(p4.x(), p4.y(), p4.z());
    glEnd();
    glBegin(GL_TRIANGLES);
    glVertex3d(p4.x(), p4.y(), p4.z());
    glVertex3d(p2.x(), p2.y(), p2.z());
    glVertex3d(p3.x(), p3.y(), p3.z());
    glEnd();
    glBegin(GL_TRIANGLES);
    glVertex3d(p1.x(), p1.y(), p1.z());
    glVertex3d(p4.x(), p4.y(), p4.z());
    glVertex3d(p3.x(), p3.y(), p3.z());
    glEnd();
#endif
    break;
  }
  post_draw(di);
}

} // End namespace SCIRun

