/*
  For more information, please see: http://software.sci.utah.edu

  The MIT License

  Copyright (c) 2009 Scientific Computing and Imaging Institute,
  University of Utah.

   
  Permission is hereby granted, free of charge, to any person obtaining a
  copy of this software and associated documentation files (the "Software"),
  to deal in the Software without restriction, including without limitation
  the rights to use, copy, modify, merge, publish, distribute, sublicense,
  and/or sell copies of the Software, and to permit persons to whom the
  Software is furnished to do so, subject to the following conditions:

  The above copyright notice and this permission notice shall be included
  in all copies or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
  OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
  THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
  DEALINGS IN THE SOFTWARE.
*/

// Core SCIRun Includes

#include <Core/Thread/Mutex.h>
#include <Core/Util/RWS.h>
#include <Core/Util/Assert.h>
#include <Core/Util/FileUtils.h>
#include <Core/Util/sci_system.h>
#include <sci_defs/version_defs.h>

// STL Includes

#include <Core/Util/Environment.h> // includes <string>
#include <Core/Util/StringUtil.h>
#include <iostream>
#include <fstream>
#include <map>
#include <list>

#define SCI_OK_TO_INCLUDE_SCI_ENVIRONMENT_DEFS_H
#include <sci_defs/environment_defs.h>

#include <stdio.h>
#ifndef _WIN32
#  include <unistd.h>
#  include <sys/param.h>
#else
#  define MAXPATHLEN 256
#  include <direct.h>
#  include <windows.h>
#  include <Core/Util/Dir.h>
#endif
#include <string.h>
#include <cctype>


#ifndef LOAD_PACKAGE
#  error You must set a LOAD_PACKAGE or life is pretty dull
#endif

#ifndef ITCL_WIDGETS
#  error You must set ITCL_WIDGETS to the iwidgets/scripts path
#endif

using namespace SCIRun;
using namespace std;


// Need these mutexes to be created here for use in another library
// as they must "construct" _now_ because they are used before the
// library that would normally construct them gets loaded...

namespace SCIRun {

SCISHARE Mutex colormapIEPluginMutex("ColorMap Import/Export Plugin Table Lock");
SCISHARE Mutex fieldIEPluginMutex("Field Import/Export Plugin Table Lock");
SCISHARE Mutex matrixIEPluginMutex("Matrix Import/Export Plugin Table Lock");
SCISHARE Mutex nrrdIEPluginMutex("Nrrd Import/Export Plugin Table Lock");

}


// This set stores all of the environemnt keys that were set when scirun was
// started. Its checked by sci_putenv to ensure we don't overwrite variables
static std::map<std::string, std::string> scirun_env;

// MacroSubstitute takes var_value returns a string with the environment
// variables expanded out.  Performs one level of substitution
//   Note: Must delete the returned string when you are done with it.
char*
MacroSubstitute( const char * var_value )
{
  int    cur = 0;
  int    start = 0;
  int    end = start;
  string newstring("");
  char * macro = 0;
  
  if (var_value==0)
    return 0;

  char* var_val = strdup(var_value);
  int length = strlen(var_val);
  while (cur < length-1) {
    if (var_val[cur] == '$' && var_val[cur+1]=='(') {
      cur+=2;
      start = cur;
      while (cur < length) {
        if (var_val[cur]==')') {
          end = cur-1;
          var_val[cur]='\0';
          macro = new char[end-start+2];
          sprintf(macro,"%s",&var_val[start]);
          const char *env = sci_getenv(macro);
          delete macro;
          if (env) 
            newstring += string(env);
          var_val[cur]=')';
          cur++;
          break;
        } else
          cur++;
      }
    } else {
      newstring += var_val[cur];
      cur++;
    }
  }

  newstring += var_val[cur]; // don't forget the last character!
  free(var_val);

  size_t newlength = strlen(newstring.c_str());
  char* retval = new char[newlength+1];
  sprintf(retval,"%s",newstring.c_str());
  
  return retval;
}

// WARNING: According to other software (specifically: tcl) you should
// lock before messing with the environment.

// Have to append 'SCIRun::' to these function names so that the
// compiler believes that they are in the SCIRun namespace (even
// though they are declared in the SCIRun namespace in the .h file...)
const char *
SCIRun::sci_getenv( const std::string & key )
{
  if (scirun_env.find(key) == scirun_env.end()) return 0;
  return scirun_env[key].c_str();
}

void
SCIRun::sci_putenv( const std::string &key, const std::string &val )
{
  scirun_env[key] = val;
}  


std::map<std::string, std::string>& 
SCIRun::get_sci_environment()
{
  return (scirun_env);
}


#ifdef _WIN32
void getWin32RegistryValues(string& obj, string& src, string& appdata, string& thirdparty, string& packages)
{
  // on an installed version of SCIRun, query these values from the registry, overwriting the compiled version
  // if not an installed version, return the compiled values unchanged
  HKEY software, company, version, scirun, pack, volatileEnvironment;
  const unsigned int PATH_SIZE = 512;
  const unsigned int NAME_SIZE = 128;
  const unsigned int MAX_VALUE_NAME = 16383;

  if (RegOpenKeyEx(HKEY_LOCAL_MACHINE, "SOFTWARE", 0, KEY_READ, &software) == ERROR_SUCCESS) {
    if (RegOpenKeyEx(software, "SCI Institute", 0, KEY_READ, &company) == ERROR_SUCCESS) {
      if (RegOpenKeyEx(company, "SCIRun", 0, KEY_READ, &scirun) == ERROR_SUCCESS) {
        if (RegOpenKeyEx(scirun, SCIRUN_VERSION, 0, KEY_READ, &version) == ERROR_SUCCESS) {
          char data[512];
          DWORD size = 512;
          DWORD type;
          int code = RegQueryValueEx(version, "InstallPath", 0, &type, (LPBYTE) data, &size);
          if (type == REG_SZ && code == ERROR_SUCCESS) {
            string installDir = data;
            convertToUnixPath(installDir);
            if (installDir[installDir.length()-1] == '/')
              installDir = installDir.substr(0, installDir.length()-1);
            obj = installDir + "/bin";
            src = installDir + "/src";
            thirdparty = installDir;
            cout << "Loading SCIRun from: " << installDir << "\n";
          }

          if (RegOpenKeyEx(version, "Packages", 0, KEY_READ|KEY_ENUMERATE_SUB_KEYS, &pack) == ERROR_SUCCESS) {
            // see http://msdn.microsoft.com/en-us/library/ms724256(VS.85).aspx
            TCHAR  name[NAME_SIZE];           // buffer for subkey name
            DWORD  nameSize;                  // size of name string 
            TCHAR  className[MAX_PATH] = TEXT("");  // buffer for class name 
            DWORD  classNameSize = MAX_PATH;  // size of class string (MAX_PATH defined in WinDef.h)
            DWORD  maxClassNameSize;          // longest class string 
            DWORD  numSubKeys = 0;            // number of subkeys 
            DWORD  maxSubKeySize;             // longest subkey size 
            DWORD  numValues;              // number of values for key 
            DWORD  maxValueName;           // longest value name 
            DWORD  maxValueData;           // longest value data 
            DWORD  securityDescriptorSize; // size of security descriptor 
            FILETIME filetime;             // last write time 

            DWORD index, code = ERROR_SUCCESS; 
            // Force SCIRun package to be first.
            // The SCIRun package must always be available.
            packages = "SCIRun,";
            bool scirunPackageFound = false;
 
            // Get the class name and the value count. 
            code = RegQueryInfoKey(pack, className, &classNameSize, 0, &numSubKeys, &maxSubKeySize,
                                         &maxClassNameSize, &numValues, &maxValueName, &maxValueData,
                                         &securityDescriptorSize, &filetime);
            // Enumerate the subkeys, until RegEnumKeyEx fails.
            if (code == ERROR_SUCCESS && numSubKeys > 0) {
              for (index = 0; index < numSubKeys; index++) {
                nameSize = NAME_SIZE;
                if (RegEnumKeyEx(pack, index, name, &nameSize, 0, 0, 0, &filetime) == ERROR_SUCCESS) {
                  if (! scirunPackageFound && string(name) == "SCIRun") {
                    scirunPackageFound = true;
                  }
                  else {
                    packages = packages + name + ",";
                  }
                }
              }
            }
            // lose trailing comma
            if (index > 0 && packages[packages.length()-1] == ',')
              packages[packages.length()-1] = 0;
            cout << "Packages: " << packages << "\n";
            RegCloseKey(pack);
          }
          RegCloseKey(version);
        }
        RegCloseKey(scirun);
      }
      RegCloseKey(company);
    }
    RegCloseKey(software);
  }

  if (RegOpenKeyEx(HKEY_CURRENT_USER, "Volatile Environment", 0, KEY_READ, &volatileEnvironment) == ERROR_SUCCESS) {
    char data[PATH_SIZE];
    DWORD size = PATH_SIZE;
    DWORD type;
    int code = RegQueryValueEx(volatileEnvironment, "APPDATA", 0, &type, (LPBYTE) data, &size);
    if (type == REG_SZ && code == ERROR_SUCCESS) {
      string userAppdata = data;
      convertToUnixPath(userAppdata);
      if (userAppdata[userAppdata.length() - 1] == '/')
        userAppdata = userAppdata.substr(0, userAppdata.length() - 1);
      appdata = userAppdata + "/SCIRun";
    }
    RegCloseKey(volatileEnvironment);
  }
}
#endif
// get_existing_env() will fill up the SCIRun::existing_env string set
// with all the currently set environment variable keys, but not their values
void
SCIRun::create_sci_environment(char **env, char *execname)
{

  // set defaults
  scirun_env.clear();
  create_environment_defaults(scirun_env);

  // override defaults
  if (env) {
    char **environment = env;
    while (*environment) {
      const string str(*environment);
      const size_t pos = str.find("=");
      scirun_env[str.substr(0,pos)] = str.substr(pos+1, str.length());
      environment++;
    }
  }

  string executable_name = "scirun";
  string objdir = SCIRUN_OBJDIR;
  string srcdir = SCIRUN_SRCDIR;
  string thirdpartydir = SCIRUN_THIRDPARTY_DIR;
  string packages = LOAD_PACKAGE;
  string iwidgets = ITCL_WIDGETS;
  string appdata;

#if defined(SCIRUN_PACKAGE_SRC_PATH)
  string pkg_src_path = SCIRUN_PACKAGE_SRC_PATH;
#endif
#if defined(SCIRUN_PACKAGE_LIB_PATH)
  string pkg_lib_path = SCIRUN_PACKAGE_LIB_PATH;
#endif

#ifdef _WIN32
  getWin32RegistryValues(objdir, srcdir, appdata, thirdpartydir, packages);
  if (!validDir(appdata)) {
    try {
      Dir d = Dir::create(appdata);
      if (!d.exists()) {
        std::cerr << "The following directory could not be created " << appdata << "\n";
      }
      else {
        if (!validDir(appdata)) {
          std::cerr << "The following directory does not exist " << appdata << "\n";
        }
      }
    }
    catch (...) {
      std::cerr << __FILE__ << ", " << __LINE__ << ": Dir::create() failed in create_sci_environment(..)" << std::endl;
    }
  }
#endif
  if (!sci_getenv("SCIRUN_OBJDIR")) 
    {
      if (!execname)
        sci_putenv("SCIRUN_OBJDIR", objdir);
      else {
        string objdir(execname);
        const char *path = sci_getenv("PATH");
        if (execname[0] != '/' && execname[1] != ':') { // unix and windows compatible
          if (objdir.find("/") == string::npos && 
              objdir.find("\\") == string::npos && 
              path) {
            objdir = findFileInPath(execname, path);
            ASSERT(objdir.length());
          } else {
            char cwd[MAXPATHLEN];
            getcwd(cwd,MAXPATHLEN);
            objdir = cwd+string("/")+objdir;
          }
        }

        string::size_type pos = objdir.find_last_of('/');
        if (pos == string::npos)
          pos = objdir.find_last_of('\\');

        executable_name = objdir.substr(pos+1, objdir.size()-pos-1);;
        objdir.erase(objdir.begin()+pos+1, objdir.end());

        sci_putenv("SCIRUN_OBJDIR",  objdir);
      }
    }

  if (!sci_getenv("SCIRUN_SRCDIR"))
    sci_putenv("SCIRUN_SRCDIR", srcdir);
  if (!sci_getenv("SCIRUN_THIRDPARTY_DIR"))
    sci_putenv("SCIRUN_THIRDPARTY_DIR", thirdpartydir);
  if (!sci_getenv("SCIRUN_LOAD_PACKAGE"))
    sci_putenv("SCIRUN_LOAD_PACKAGE", packages);
  if (!sci_getenv("SCIRUN_ITCL_WIDGETS"))
    sci_putenv("SCIRUN_ITCL_WIDGETS", iwidgets);
  
#if defined(SCIRUN_PACKAGE_SRC_PATH)
  if(!sci_getenv("SCIRUN_PACKAGE_SRC_PATH") &&  pkg_src_path != "")
    sci_putenv("SCIRUN_PACKAGE_SRC_PATH", pkg_src_path);
#endif
#if defined(SCIRUN_PACKAGE_LIB_PATH)
  if(!sci_getenv("SCIRUN_PACKAGE_LIB_PATH") && pkg_lib_path != "")
    sci_putenv("SCIRUN_PACKAGE_LIB_PATH", pkg_lib_path);
#endif

  if (!sci_getenv("SCIRUN_TMP_DIR"))
    sci_putenv("SCIRUN_TMP_DIR", "/tmp/");

  // native windows doesn't have "HOME"
  if (!sci_getenv("HOME")) {
    // Use appropriate AppData directory (VISTA permissions) instead of OBJTOP (previous versions did this)
    sci_putenv("HOME", appdata);
  }

  sci_putenv("SCIRUN_ITCL_WIDGETS", 
             MacroSubstitute(sci_getenv("SCIRUN_ITCL_WIDGETS")));

  sci_putenv("EXECUTABLE_NAME", executable_name);
  string rcfile = "." + executable_name + "rc";
  find_and_parse_rcfile(rcfile);
}

// emptryOrComment returns true if the 'line' passed in is a comment
// ie: the first non-whitespace character is a '#'
// or if the entire line is empty or white space.
bool
emptyOrComment( const char * line )
{
  const char A_TAB = 9; // 9 is the ascii value of TAB
  int   length = static_cast<int>( strlen( line ) );

  for( int pos = 0; pos < length; pos++ ) {
    if( line[pos] == '#' ) {
      return true;
    } else if( ( line[pos] != ' ' ) && ( line[pos] != A_TAB ) ) {
      return false;
    }
  }
  return true;
}

/// Tests a string to see if it is devoid of content.
static bool
str_vacuous(const char *s)
{
  const char *iter = s;
  while(*iter) {
    if(!isspace(*iter)) {
      return false;
    }
    ++iter;
  }
  return true;
}

// parse_rcfile reads the file 'rcfile' into SCIRuns enviroment mechanism
// It uses sci_putenv to set variables in the environment. 
// Returns true if the file was opened and parsed.  False otherwise.
bool
SCIRun::parse_rcfile( const string &rcfile )
{
  if (rcfile.size() == 0) return (false);
 
  if ( !validFile( rcfile )) return (false);
    
  std::ifstream file;

  try
  {
    file.open(rcfile.c_str());
  }
  catch(...)
  {
    return (false);
  }

  std::string line;
  
  while(!file.eof())
  { 
    std::getline(file,line);

    std::string key;
    std::string value;
    
    size_t startloc = 0;
    while (startloc < line.size() && ((line[startloc] == ' ')||(line[startloc] == '\t'))) startloc++;
    size_t equalloc = line.find("=");
    if (equalloc != std::string::npos)
    {
      std::string key = line.substr(startloc,equalloc-startloc);
      size_t endloc = equalloc-startloc-1;
      while (endloc > 0 && ((key[endloc] == ' ')||(key[endloc] == '\t')) ) endloc--;
      key = key.substr(0,endloc+1);

      value = line.substr(equalloc+1);
      startloc = 0;
      while (startloc < value.size() && ((value[startloc] == ' ')||(value[startloc] == '\t'))) startloc++;
      value = value.substr(startloc);

      endloc = value.size()-1;
      while (endloc > 0 && ((value[endloc] == ' ')||(value[endloc] == '\t')||
              (value[endloc] == '\n')||(value[endloc] == '\r')) ) endloc--;

      value = value.substr(0,endloc+1);

      if (key.size() > 0 && key[0] != '#')
      {
        sci_putenv(key,value);
      }
    }


  }
  file.close();

  sci_putenv("SCIRUN_RC_PARSED","1");
  sci_putenv("SCIRUN_RCFILE",rcfile);
  
  return true;
}

// find_and_parse_rcfile will search for the rcfile file in 
// default locations and read it into the environemnt if possible.
void
SCIRun::find_and_parse_rcfile(const string &rcfile)
{
  bool foundrc=false;
  const string slash("/");

  // 1. check the local directory
  string filename(rcfile);
  foundrc = parse_rcfile(filename);
  
  // 2. check the BUILD_DIR
  if (!foundrc) {
    filename = SCIRUN_OBJDIR + slash + string(rcfile);
    foundrc = parse_rcfile(filename);
  }
  
  // 3. check the user's home directory
  const char *HOME = NULL;
  if (!foundrc && (HOME = sci_getenv("HOME"))) {
    filename = HOME + slash + string(rcfile);
    foundrc = parse_rcfile(filename);
  }
  
  // 4. check the source code directory
  if (!foundrc) {
    filename = SCIRUN_SRCDIR + slash + string(rcfile);
    foundrc = parse_rcfile(filename);
  }

  if (foundrc) sci_putenv("SCIRUN_RCFILE",filename);

  // Tell the user that we parsed the the rcfile...
//  std::cout << "Parsed " << rcfile << "... " << filename << "\n";
}


bool
SCIRun::update_rcfile(std::string key, std::string value)
{
  std::string filename = sci_getenv("SCIRUN_RCFILE");
  if (filename.size() == 0) return (false);
  
  std::ifstream file;

  try
  {
    file.open(filename.c_str());
  }
  catch(...)
  {
    return (false);
  }

  std::string line;
  std::list<std::string> lines;
  
  bool found = false;
        
  while(!file.eof())
  { 
    std::getline(file,line);
    size_t startloc = 0;
    while (startloc < line.size() && ((line[startloc] == ' ')||(line[startloc] == '\t'))) startloc++;
    size_t equalloc = line.find("=");
    if (equalloc != std::string::npos)
    {
      std::string keyname = line.substr(startloc,equalloc-startloc);
      size_t endloc = equalloc-startloc-1;
      while (endloc > 0 && ((keyname[endloc] == ' ')||(keyname[endloc] == '\t')) ) endloc--;
      keyname = keyname.substr(0,endloc+1);
      if ((keyname == key) || (keyname == ("# "+key)))
      {
        line = key + " = " + value;
        found = true;
      }
    }
    line += "\n";
    lines.push_back(line);
  }
  file.close();
  
  if (found == false)
  {
    line = key + "=" + value + "\n";
    lines.push_back(line);
  }
  
  std::ofstream nfile;
  
  try
  {
    nfile.open(filename.c_str());
  }
  catch(...)
  {
    return (false);
  }

  std::list<std::string>::iterator it, it_end;
  it = lines.begin();
  it_end = lines.end();
  
  while(it != it_end)
  {
    nfile << (*it); ++it;
  }

  nfile.close();
  return (true);
}


void
SCIRun::copy_and_parse_scirunrc()
{
  try {
    string home   = sci_getenv("HOME");
    string srcdir = sci_getenv("SCIRUN_SRCDIR");

    if( home == "" || srcdir == "" ) {
      cout << "Warning: HOME ('" << home << "') or SCIRUN_SRCDIR ('" << srcdir << "') invalid... skipping copy...\n";
      return;
    }

    string srcRc = srcdir + "/scirunrc";

    string homerc = home + "/.scirunrc";
    if (validFile(homerc)) {
      const char* env_rcfile_version = sci_getenv("SCIRUN_RCFILE_VERSION");
      string backup_extension =(env_rcfile_version ? env_rcfile_version:"bak");
      string backuprc = homerc + "." + backup_extension;

      std::cout << "Backing up " << homerc << " to " << backuprc << endl;
      copyFile(homerc, backuprc);
    }

    std::cout << "Copying " << srcRc << " to " << homerc << "\n";
    if (copyFile(srcRc, homerc) == 0) 
    {
      // If the scirunrc file was copied, then parse it.
      parse_rcfile(homerc);
    }
  }
  catch (...) {
    std::cerr << __FILE__ << ", " << __LINE__ << ": copy_and_parse_scirunrc() failed" << std::endl;
  }
}

// sci_getenv_p will lookup the value of the environment variable 'key' and 
// returns false if the variable is equal to 'false', 'no', 'off', or '0'
// returns true otherwise.  Case insensitive.
bool
SCIRun::sci_getenv_p(const string &key) 
{
  const char *value = sci_getenv(key);

  // If the environment variable does NOT EXIST OR is EMPTY then return FASE
  if (!value || !(*value)) return false;
  string str;
  while (*value) {
    str += toupper(*value);
    value++;
  }

  // Only return false if value is zero (or equivalant)
  if (str == "FALSE" || str == "NO" || str == "OFF" || str == "0")
    return false;
  // Following C convention where any non-zero value is assumed true
  return true;
}

bool
SCIRun::replace_environment_variables(std::string& str)
{
  // scan for variable names
  
  std::string::size_type size = str.size();
  std::string::size_type old_size = str.size();
  std::string::size_type start_loc = 0;
  std::string::size_type dollar_loc = str.find('$',start_loc);
  
  while (dollar_loc != std::string::npos)
  {
    // found a possible variable
    if (dollar_loc+3 < size && str[dollar_loc+1] == '{')
    {
      std::string::size_type end_brac = str.find('}',dollar_loc+1);
      if (end_brac != std::string::npos)
      {
        std::string key = str.substr(dollar_loc+2,end_brac-dollar_loc-2);
        if (sci_getenv(key))
        {
          std::string value = sci_getenv(key);
          old_size = size;
          str = str.substr(0,dollar_loc)+value+str.substr(end_brac+1);
          size = str.size();
          start_loc = size - end_brac;
        }
      }
    }
    else
    {
      std::string::size_type end_loc = start_loc+1;
      
      while ( (end_loc < size) && ((str[end_loc] >= 'a' && str[end_loc] <= 'z')||
              (str[end_loc] >= 'A' && str[end_loc] <= 'Z') || 
              (str[end_loc] >= '0' && str[end_loc] <= '9') ||
              (str[end_loc] == '_' ))) end_loc++;
              
      if (end_loc > start_loc+1)
      {
        std::string key = str.substr(dollar_loc+1,end_loc-dollar_loc-1);
        if (sci_getenv(key))
        {
          std::string value = sci_getenv(key);
          old_size = size;
          str = str.substr(0,dollar_loc)+value+str.substr(end_loc);
          size = str.size();
          start_loc = size - end_loc - 1;
        }        
      }
    }
    dollar_loc = str.find('$',start_loc+1);
    start_loc++;
  }
  return (true);
}



