/*
 * The MIT License
 * 
 * Copyright (c) 2009 Scientific Computing and Imaging Institute,
 * University of Utah.
 * 
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 */

#include <Core/Datatypes/FieldArray.h>

#include <Core/Algorithms/Fields/TransformMesh/GeneratePolarProjectionImage.h>
#include <Core/Algorithms/Fields/Mapping/MapFromPointField.h>
#include <Core/Algorithms/Fields/Mapping/FindClosestNodesFromPointField.h>

#include <Core/Datatypes/Matrix.h>
#include <Core/Datatypes/DenseMatrix.h>

#include <Dataflow/Widgets/ArrowWidget.h>
#include <Core/Thread/CrowdMonitor.h>

#include <Dataflow/Network/Ports/GeometryPort.h>
#include <Dataflow/Network/Ports/FieldPort.h>
#include <Dataflow/Network/Ports/MatrixPort.h>
#include <Dataflow/Network/Module.h>

namespace SCIRun {

using namespace SCIRun;

class GeneratePolarProjectionOnImageForFieldArray : public Module {
  public:
    GeneratePolarProjectionOnImageForFieldArray(GuiContext*);
    virtual ~GeneratePolarProjectionOnImageForFieldArray();

    virtual void execute();

    virtual void widget_moved(bool release, BaseWidget*);

  private:  
    SCIRunAlgo::GeneratePolarProjectionImageAlgo algo_;
    SCIRunAlgo::MapFromPointFieldAlgo palgo_;
    SCIRunAlgo::FindClosestNodesFromPointFieldAlgo falgo_;
    
    ArrowWidget*             widget_;
    CrowdMonitor             widget_lock_;   
    GeomHandle               widget_switch_;
    int                      widget_id_;
    
    GuiDouble                dx_,dy_,dz_,px_,py_,pz_;
    
    GuiInt                   xsize_,ysize_;
    
    Point                    origin_;
    Vector                   direction_;

    Point                    old_origin_;
    Vector                   old_direction_;
};


DECLARE_MAKER(GeneratePolarProjectionOnImageForFieldArray)

GeneratePolarProjectionOnImageForFieldArray::GeneratePolarProjectionOnImageForFieldArray(GuiContext* ctx) :
  Module("GeneratePolarProjectionOnImageForFieldArray", ctx, Source, "FieldArray", "SCIRun"),
  widget_lock_("Arrow Widget Lock"),
  dx_(get_ctx()->subVar("dx"),0.0),
  dy_(get_ctx()->subVar("dy"),0.0),
  dz_(get_ctx()->subVar("dz"),10.0),
  px_(get_ctx()->subVar("px"),0.0),
  py_(get_ctx()->subVar("py"),0.0),
  pz_(get_ctx()->subVar("pz"),0.0),
  xsize_(get_ctx()->subVar("x-size"),128),  
  ysize_(get_ctx()->subVar("y-size"),128)  
{
  algo_.set_progress_reporter(this);
  palgo_.set_progress_reporter(this);
  falgo_.set_progress_reporter(this);
  
  
  get_ctx()->reset();

  widget_ = new ArrowWidget(this, &widget_lock_, 1.0);

  widget_->SetScale(5);
  widget_->SetLength(25);
  double dx = dx_.get();
  double dy = dy_.get();
  double dz = dz_.get();
  double px = px_.get();
  double py = py_.get();
  double pz = pz_.get();
  
  widget_->SetPosition(Point(px,py,pz));
  widget_->SetDirection(Vector(dx,dy,dz));
  widget_switch_ = widget_->GetWidget();
  
  GeometryOPortHandle geom_oport;
  get_oport_handle("ProjectionVector",geom_oport);
  
  ((GeomSwitch*)widget_switch_.get_rep())->set_state(1);
  widget_id_ = geom_oport->addObj(widget_switch_,"ProjectionVector",&widget_lock_);
  geom_oport->flushViews();
  
  origin_ = Point(px,py,pz);
  direction_ = Vector(dx,dy,dz);

  old_direction_ = direction_;
  old_origin_ = origin_;
}


GeneratePolarProjectionOnImageForFieldArray::~GeneratePolarProjectionOnImageForFieldArray()
{
  delete widget_;
}

void
GeneratePolarProjectionOnImageForFieldArray::execute()
{
  // Troubles in setup, since initialization runs through TCL....
  // Cannot access these in constructor hence need to do them here
  double dx = dx_.get();
  double dy = dy_.get();
  double dz = dz_.get();
  double px = px_.get();
  double py = py_.get();
  double pz = pz_.get();
  
  widget_->SetPosition(Point(px,py,pz));
  widget_->SetDirection(Vector(dx,dy,dz));
  origin_ = Point(px,py,pz);
  direction_ = Vector(dx,dy,dz);

  FieldArrayHandle input, output;
  FieldHandle geometry;
  get_input_handle("FieldArray",input,true);
  get_input_handle("Geometry",geometry,true);
  
  MatrixHandle origin, direction;
  get_input_handle("Origin",origin,false);
  get_input_handle("Direction",direction,false);
  
  if (origin.get_rep() && origin->get_data_size() > 2)
  {
    double *data = origin->get_data_pointer();
    origin_ = Point(data[0],data[1],data[2]);
    widget_->SetPosition(origin_);
  }
  
  if (direction.get_rep() && direction->get_data_size() > 2)
  {
    double *data = direction->get_data_pointer();
    direction_ = Vector(data[0],data[1],data[2]);
    widget_->SetDirection(direction_);
  }
  
  if (inputs_changed_ || origin_ != old_origin_ ||
      xsize_.changed() || ysize_.changed() ||
      direction_ != old_direction_ || !oport_cached("ProjectedData") ||
      !oport_cached("ProjectedPoints") || !oport_cached("ProjectedArea") ||
      !oport_cached("ProjectedMask") ||
      !oport_cached("Origin") || !oport_cached("Direction"))
  {
    update_state(Executing);

    int xsize = xsize_.get();
    int ysize = ysize_.get();
    
    algo_.set_point("origin",origin_);
    algo_.set_vector("direction",direction_);
    
    algo_.set_int("xsize",xsize);
    algo_.set_int("ysize",ysize);
    
    FieldHandle joinedfield, projecteddata, projectedpoints, projectedmask, projectedarea;

    std::vector<FieldHandle>& inputs = input->array();
    
    if(!(algo_.run(geometry,projectedpoints,projectedmask,projectedarea))) return;
    
    size_t size = inputs.size();

    geometry->vmesh()->synchronize(Mesh::EPSILON_E);
    palgo_.set_scalar("max_distance",10*geometry->vmesh()->get_epsilon());
    
    
    FieldInformation fi(LATVOLMESH_E,CONSTANTDATA_E,DOUBLE_E);
    MeshHandle mesh = CreateMesh(fi,xsize,ysize,size+1,Point(0.0,0.0,0.0),Point(1.0,1.0,1.0));
    Transform trans;
    projectedmask->vmesh()->get_canonical_transform(trans);
    mesh->vmesh()->transform(trans);

    FieldHandle output = CreateField(fi,mesh);
    
    Field::size_type sz = (xsize-1)*(ysize-1);
    for (size_t j=0; j<size; j++)
    {
      FieldHandle data;
      if(!(palgo_.run(inputs[j],projectedpoints,data))) return;
      if (inputs[0]->mesh().get_rep() != inputs[j]->mesh().get_rep()) inputs[j]->vmesh()->clear_synchronization(); // to save memory  
      output->vfield()->copy_values(data->vfield(),0,j*sz,sz);
      update_progress(j,size);
    }
    
    FieldHandle projectedindex;
    if(!(falgo_.run(geometry,projectedpoints,projectedindex))) return;
    

    old_origin_ = origin_;
    old_direction_ = direction_;
    send_output_handle("ProjectedData",output,true);
    send_output_handle("ProjectedPoints",projectedpoints,true);
    send_output_handle("ProjectedMask",projectedmask,true);
    send_output_handle("ProjectedArea",projectedarea,true);
    send_output_handle("ProjectedIndex",projectedindex,true);
    
    origin = new DenseMatrix(3,1);
    double* data = origin->get_data_pointer();
    data[0] = origin_.x(); data[1] = origin_.y(); data[2] = origin_.z();
    send_output_handle("Origin",origin,true);

    direction = new DenseMatrix(3,1);
    data = direction->get_data_pointer();
    data[0] = direction_.x(); data[1] = direction_.y(); data[2] = direction_.z();
    send_output_handle("Direction",direction,true);    
  }
}

void
GeneratePolarProjectionOnImageForFieldArray::widget_moved(bool release, BaseWidget*)
{
  if (release) 
  {
    Point  p = widget_->GetPosition();
    Vector d = widget_->GetDirection();
    
    px_.set(p.x()); py_.set(p.y()); pz_.set(p.z());
    dx_.set(d.x()); dy_.set(d.y()); dz_.set(d.z());
    
    want_to_execute();
    origin_ = p;
    direction_ = d;
  }
}

} // End namespace SCIRun


