/*
 * The MIT License
 * 
 * Copyright (c) 2008 Scientific Computing and Imaging Institute,
 * University of Utah.
 * 
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 */


#include <Dataflow/Network/Module.h>
#include <Dataflow/Network/Ports/GeometryPort.h>
#include <Dataflow/Network/Ports/FieldPort.h>
#include <Dataflow/Network/Ports/MatrixPort.h>
#include <Core/Datatypes/Field.h>
#include <Core/Datatypes/Mesh.h>
#include <Core/Datatypes/CurveMesh.h>
#include <Core/Datatypes/MeshTypes.h>
#include <Core/Datatypes/Matrix.h>
#include <Core/Geometry/Point.h>
#include <Core/Datatypes/ColumnMatrix.h>
#include <Core/Datatypes/SparseRowMatrix.h>
#include <Core/Datatypes/DenseMatrix.h>
#include <Core/Datatypes/PointCloudMesh.h>
#include <Core/Datatypes/VMesh.h>

#include <Core/Thread/CrowdMonitor.h>
#include <Dataflow/Widgets/PointWidget.h>

#include <iostream>
#include <list>
#include <string>
#include <stack>


namespace SCIRun {
	
	using std::stack;

	// equialent to the interp1 command in matlab.  uses the parameters p and t to perform a cubic spline interpolation pp in one direction.
	
	bool
	CalculateSpline(std::vector<double>& t, std::vector<double>& x, std::vector<double>& tt, std::vector<double>& xx)
	{
		// need to have at least 3 nodes
		if (t.size() < 3) return (false);
		if (x.size() != t.size()) return (false);
		
		
		//cout<<"------beginning spline algorithm"<<endl;
		
		
		size_t size = x.size();
		vector<double> z(size), h(size-1), b(size-1), v(size-1), u(size-1);
		
		for (size_t k=0;k<size-1;k++)
		{
			h[k]=(t[k+1]-t[k]);
			b[k]=(6*(x[k+1]-x[k])/h[k]);
			
			//cout<<"--- h="<<h[k]<<", b="<<b[k]<<"  and k = "<< k<< endl;
		}
		
		//cout<<"-----h and b  calculated"<<endl;
		
		
		u[1]=2*(h[0]+h[1]);
		v[1]=b[1]-b[0];
		
		for (size_t k=2;k<size-1;k++)
		{
			u[k]=2*(h[k]+h[k-1])-(h[k-1]*h[k-1])/u[k-1];
			v[k]=b[k]-b[k-1]-h[k-1]*v[k-1]/u[k-1];
			
			//cout<<"--- u="<<u[k]<<", v="<<v[k]<<"  and k = "<< k<< endl;
		}
		
		//cout<<"----- u and v calculated"<<endl;
		
		z[size-1]=0;
		
		for (size_t k=size-2;k>0;k--)
		{
			z[k]=(v[k]-h[k]*z[k+1])/u[k];
			
			//cout<<"--- z="<<z[k]<<"  and k = "<< k<< endl;
			
		}
		
		z[0]=0;
		
		//cout<<"----- z calculated"<<endl;
		
		size_t segment = 0;
		
		xx.resize(tt.size());
		for(size_t k = 0; k < tt.size(); k++)
		{
			while (segment < (size-2) && t[segment+1] < tt[k])
			{
				segment++;
				//cout<<"----- segment number "<<segment<<endl;
			}
			
			
			
			double w0,w1,w2,w3,a,b,c,d;
			
			w0=(t[segment+1]-tt[k])*(t[segment+1]-tt[k])*(t[segment+1]-tt[k]);
			w1=(tt[k]-t[segment])*(tt[k]-t[segment])*(tt[k]-t[segment]);
			w2=(tt[k]-t[segment]);
			w3=(t[segment+1]-tt[k]);
			
			a=z[segment]/(6*h[segment]);
			b=z[segment+1]/(6*h[segment]);
			c=(x[segment+1]/h[segment]-(z[segment+1]*h[segment])/6);
			d=(x[segment]/h[segment]-z[segment]*h[segment]/6);
					
			xx[k]=a*w0+b*w1+c*w2+d*w3;
			
		}
		
		//cout<<"----- xx calculated"<<endl;
		
		return (true);
	}

	
	
	// this is a sline function.  pp is the final points that are in between the original points p.  
	// t and tt are the original and final desired spacing, respectively.
	
	bool
	CalculateSpline(std::vector<double>& t, std::vector<Point>& p, std::vector<double>& tt, std::vector<Point>& pp)
	{
		// need to have at least 3 nodes
		if (t.size() < 3) return (false);
		if (p.size() != t.size()) return (false);
		
		
		size_t size=p.size();
		
		std::vector<double> x(size), y(size), z(size);
		std::vector<double> xx, yy, zz;
		
		
		
		for (size_t	k=0;k<p.size();k++)
		{
			x[k]=p[k].x();
			y[k]=p[k].y();
			z[k]=p[k].z();
		}
		
		//cout<< "-----widget points reassigned"<< endl;
		
		CalculateSpline(t,x,tt,xx);
		CalculateSpline(t,y,tt,yy);
		CalculateSpline(t,z,tt,zz);
		
		//cout<< "-----executed interpolation commands"<< endl;
		
		
		for (size_t	k=0;k<tt.size();k++)
		{
			pp.push_back(Point(xx[k],yy[k],zz[k]));
					
		}
		
		//cout<< "----spline interpolation  done!!"<< endl;
		
		return (true);
	}
	
	class GenerateWireElectrode : public Module
		{
		public:
			GenerateWireElectrode(GuiContext* ctx);
			~GenerateWireElectrode(){};
			virtual void execute();
			virtual void widget_moved(bool, BaseWidget*);
			virtual void tcl_command(GuiArgs& args, void* userdata);
			
			virtual void post_read(); // get the widget state...
			virtual void presave();
			
			void add_point(vector<Point>& p);
			bool remove_point();
			
			//void reset();
			
			
			void create_widgets(std::vector<Point>& points);
			
			void get_points(std::vector<Point>& points);
			void get_centers(std::vector<Point>& p,std::vector<Point>& pp);
			
			
		private:
			
			CrowdMonitor										widget_lock_;
			GeomHandle											widget_switch_;
			GeometryOPortHandle							geom_oport_;
			
			std::vector<PointWidget*>				widget_;
			
			std::vector<GuiPointHandle>			widget_point_;
			GuiInt													gui_widget_points_;
			
			GuiDouble												gui_length_;
			GuiDouble												gui_width_;
			
			bool														color_changed_;
			
			GuiInt													gui_use_field_;
		
			GuiDouble												gui_color_r_;
			GuiDouble												gui_color_g_;
			GuiDouble												gui_color_b_;
			
			GuiDouble												gui_probe_scale_;
			
			GuiString												gui_moveto_;
			
			GuiInt													gui_wire_res_;
			
						
	
			
		};
	
	
	
	DECLARE_MAKER(GenerateWireElectrode)
	
	GenerateWireElectrode::GenerateWireElectrode(GuiContext* ctx) 
		:	Module("GenerateWireElectrode", ctx, Source, "NewField", "SCIRun"),
		widget_lock_("GenerateWireElectrode widget lock"),
		gui_probe_scale_(get_ctx()->subVar("probe_scale"), 3.0),
		gui_color_r_(get_ctx()->subVar("color-r"), 1.0),
		gui_color_g_(get_ctx()->subVar("color-g"), 1.0),
		gui_color_b_(get_ctx()->subVar("color-b"), 1.0),
		gui_widget_points_(get_ctx()->subVar("num_points"),5),
		gui_length_(get_ctx()->subVar("length"),100),
		gui_width_(get_ctx()->subVar("width"),5),
		gui_moveto_(get_ctx()->subVar("moveto", false), ""),
		gui_use_field_(get_ctx()->subVar("use-field"),1),
		gui_wire_res_(get_ctx()->subVar("wire_res"),10),
		color_changed_(false)
	
	
	
		{
			get_oport_handle("GenerateWireElectrode Widget",geom_oport_);
		}
	
	void
	GenerateWireElectrode::execute()
	{
		//cout<< "----begin execute" << endl;
		
		FieldHandle source, ofield;
		const bool input_field_p =get_input_handle("Input Field",source,false);
		
		//if (inputs_changed_ || !oport_cached("Output Field") )
		//{
			update_state(Executing);
				
			//size_t num_points=gui_widget_points_.get();
		
		//cout<<"--- beginning widget size="<<widget_.size()<<".  num_points= "<< num_points<<endl;
		
			
			FieldInformation fis(source);
			
			
			vector<Point> orig_points;
		
		
			const string &moveto = gui_moveto_.get();
		
			//cout<<"moveto="<<moveto<<".   number of points: "<<num_points<<endl;
					
			double Pi=3.14159;
		
			int use_field=gui_use_field_.get();
		
		
		
		
			if (input_field_p && (use_field==1) && (moveto=="default"))
			{
				//cout<< "----using field data"<< endl;
				VMesh* smesh = source->vmesh();				
				VField* sfield = source->vfield();
				
				smesh->synchronize(Mesh::ELEM_LOCATE_E);
				
				VMesh::Node::size_type num_nodes = smesh->num_nodes();
				if (num_nodes>50)
				{
					error("Why would you want to use that many nodes to make an electrode?  Do you want to crash you system?  That's way to many.");
					return;
				}
				
				VMesh::Node::array_type a;
				
				orig_points.resize(num_nodes);
				
				for (VMesh::Node::index_type idx=0;idx<num_nodes;idx++)
				{
					
					Point ap;
					smesh->get_center(ap,idx);
					
					orig_points[idx]=ap;
				}
				
				gui_moveto_.set("");
				
									
			}
			
			else if ((!input_field_p || use_field==0) && (moveto=="default" || widget_.size()==0))
			{
				double l, lx;
				l=gui_length_.get();
				
				lx=l*.5774;
				
				orig_points.resize(5);
				
				orig_points[0]=(Point(0,0,0));
				orig_points[1]=(Point(lx*.25,lx*.25,lx*.25));
				orig_points[2]=(Point(lx*.5,lx*.5,lx*.5));
				orig_points[3]=(Point(lx*.75,lx*.75,lx*.75));
				orig_points[4]=(Point(lx,lx,lx));
				
											
				//cout<<"----using default positions"<<endl;
				
				gui_moveto_.set("");

				
			}
							
			else if (moveto=="add_point")
			{
				add_point(orig_points);
				
				gui_moveto_.set("");
				
				//return;
			}
			
			else if (moveto=="remove_point")
			{
				remove_point();
				
				gui_moveto_.set("");
				
				return;
			}
		

							 
			else
			{	
				
				//cout<< "----  case 3"<< endl;
				
				orig_points.resize(widget_.size());
				
				for (size_t k = 0; k < widget_.size(); k++)
				{
					orig_points[k] = widget_[k]->GetPosition();
					//cout<<"---- position "<<k<<" = "<<orig_points[k]<<endl;
				}
				//cout<<"---using widget position"<<endl;
				
				
			}
		
			gui_widget_points_.set(orig_points.size());
			 
			
			FieldInformation fi("TetVolMesh",0,"double");
			MeshHandle mesh = CreateMesh(fi);
			VMesh::Node::array_type nodes;
			 
		
				
		//cout<<"--- original point vector size="<<orig_points.size()<<endl;

			
		
			size_type size=orig_points.size();
			vector<Point> final_points;
			
			vector<Point> points(size);
			
				
			create_widgets(orig_points);		
			
			get_centers(points,final_points);
	
			double radius=gui_width_.get();
			
			size_t DN=gui_wire_res_.get();
			
			Vector Vold1, Vold2;
			Vector V1, V2, V, Vx, Vy;
			
			Vold1[0]=1;
			Vold1[1]=0;
			Vold1[2]=0;
			
			Vold2[0]=0;
			Vold2[1]=1;
			Vold2[2]=0;
			
			size_t N=final_points.size();
			
			vector<Point> p=final_points;
			
			vector<double> phi(DN);
			
			vector<Point> fin_nodes;
			
			for (size_t q=0;q<DN;q++) 
			{
				phi[q]=q*(2*Pi/(static_cast<double> (DN))); 
				//cout <<"q=  "<<q<< ".  phi= "<<phi[q]<<". DN= "<<DN<<endl;
			}
		
			for (size_t k=0;k<N;k++)
			{
				
				if (k==N-1)
				{
					V1=p[k]-p[k-1];
					V2=V1;
				}
				
				else if (k==0)
				{
					V2=p[k+1]-p[k];
					V1=V2;
				}
				else
				{
					V1=p[k]-p[k-1];
					V2=p[k+1]-p[k];
				}
				
				
				if (sqrt(V1[0]*V1[0] + V1[1]*V1[1] + V1[2]*V1[2])>0) 
				{
					V1.normalize();
				}
				else
				{
					V1[0]=1; V1[1]=0; V1[2]=0; 
				}
				
				if (sqrt(V2[0]*V2[0] + V2[1]*V2[1] + V2[2]*V2[2])>0) 
				{
					V2.normalize();
				}
				else
				{
					V2[0]=0; V2[1]=1; V2[2]=0;
				}
				
				
				V=(V1+V2)*.5;
				V.normalize();
				
				//cout<<"k=  "<<k<<".  V1= "<<V1<<".  V2=  "<<V2<<". V=  "<<V<<endl;
				
				
				if (Dot(V,Vold1)<.9)
				{
					Vx=Cross(V,Vold1);
					Vy=Cross(V,Vx);
				}
				else 
				{
					Vx=Cross(V,Vold2);
					Vy=Cross(V,Vx);
				}
				
				
				Vx.normalize();
				Vy.normalize();
				
				Vold1=-Vy;
				Vold2=Vx;
				
				fin_nodes.push_back(p[k]);
				
				for (size_t q=0;q<DN;q++)
				{
					fin_nodes.push_back(Point(p[k]+Vx*radius*cos(phi[q])+Vy*radius*sin(phi[q])));
					
					//cout<<"last Point: "<<fin_nodes[k]<<".  q=  "<<q<<endl;
					
				}
				
									
			}
		
			
			
			for (VMesh::Node::index_type idx=0;idx<fin_nodes.size();idx++)
			{
				
				mesh->vmesh()->add_point(fin_nodes[idx]);
				mesh->vmesh()->get_nodes(nodes,idx);
							
			}
		
			VMesh::Node::index_type SE=0, EE=0, SE1, SE2, EE1, EE2;
		
			vector<size_t>  L(DN+1);
		
			for (size_t k=0;k<DN;k++)
			{
				L[k]=k+1;
			}
		L[DN]=1;
		
			for (VMesh::Node::index_type idx=0;idx<final_points.size()-1;idx++)
			{
				SE=static_cast<VMesh::Node::index_type> (static_cast<int> (SE)+(static_cast<int> (DN))+1);
				
				for (VMesh::Node::index_type k=0;k<DN;k++)
				{
					SE1=static_cast<VMesh::Node::index_type> (static_cast<int> (SE)+(static_cast<int> (L[k])));
					SE2=static_cast<VMesh::Node::index_type> (static_cast<int> (SE)+(static_cast<int> (L[k+1])));
					EE1=static_cast<VMesh::Node::index_type> (static_cast<int> (EE)+(static_cast<int> (L[k])));
					EE2=static_cast<VMesh::Node::index_type> (static_cast<int> (EE)+(static_cast<int> (L[k+1])));
					
					VMesh::Node::array_type elem_nodes(4);
					
					elem_nodes[0]=EE;
					elem_nodes[1]=EE1;
					elem_nodes[2]=EE2;
					elem_nodes[3]=SE;
					
					mesh->vmesh()->add_elem(elem_nodes);
					
					elem_nodes[0]=EE1;
					elem_nodes[1]=EE2;
					elem_nodes[2]=SE2;
					elem_nodes[3]=SE;
					
					mesh->vmesh()->add_elem(elem_nodes);
					
					elem_nodes[0]=EE1;
					elem_nodes[1]=SE1;
					elem_nodes[2]=SE2;
					elem_nodes[3]=SE;
					
					mesh->vmesh()->add_elem(elem_nodes);
				}
				
				EE=static_cast<VMesh::Node::index_type> (static_cast<int> (EE)+(static_cast<int> (DN))+1);
				
				
			}
		
		
		
			//add_edge(nodes);
		
			fi.make_double();
			ofield = CreateField(fi,mesh);				
			
			send_output_handle("Output Field",ofield);
		
		//cout<<"--- end widget size="<<widget_.size()<<endl;
			
				
			
			
						
						
		//}
		
		
		
		
				
	}
	
	void
	GenerateWireElectrode::widget_moved(bool release, BaseWidget* widget_)
	{
		if (release) want_to_execute();
	}
	
	
	void
	GenerateWireElectrode::create_widgets(std::vector<Point>& points)
	{
		GeomGroup *group = new GeomGroup;
		widget_switch_ = new GeomSwitch(group);
		
		geom_oport_->delAll();
		widget_.clear();
		
		double scale;
		scale=gui_probe_scale_.get()*gui_width_.get()*.5;
		
		
		for (size_t k = 0; k < points.size(); k++)
		{
			PointWidget *widget =  new PointWidget(this, &widget_lock_,gui_probe_scale_.get());
			widget_.push_back(widget);
					
			widget->Connect(geom_oport_.get_rep());
			widget->SetCurrentMode(0);
			widget->SetScale(scale);
			widget->SetPosition(points[k]);
			widget->SetColor(Color(gui_color_r_.get(),gui_color_g_.get(),gui_color_b_.get()));
			
			group->add(widget->GetWidget().get_rep());
		}
		
		geom_oport_->addObj(widget_switch_,"Wire Electrode", &widget_lock_);
	}
	
	
	void
	GenerateWireElectrode::get_points(std::vector<Point>& points)
	{
		points.resize(widget_.size());
		for (size_t k = 0; k < widget_.size(); k++)
		{
			points[k] = widget_[k]->GetPosition();
		}
	}
	
	
	
	void
	GenerateWireElectrode::get_centers(std::vector<Point>& p, std::vector<Point>& pp)
	{
		get_points(p);
		
		std::vector<double> t(p.size());

		
		t[0]=0;
		
		for (size_t k=1; k<p.size(); k++)
		{
			t[k] = (p[k]-p[k-1]).length() + t[k-1];		
		}
		
		double length=gui_length_.get();
		
		double res=gui_wire_res_.get();
		
		std::vector<double> tt(res*(p.size()-1));
		for (size_t k=0; k< tt.size(); k++) tt[k] = static_cast<double>(k)*(length/(static_cast<double>(tt.size()-1)));
		
		CalculateSpline(t,p,tt,pp);
					
				
	} 
		
			
	void 
	GenerateWireElectrode::tcl_command(GuiArgs& args, void* userdata)
	{
		if(args.count() < 2) 
		{
			args.error("ShowString needs a minor command");
			return;
		}
		
		if (args[1] == "color_change") 
		{
			color_changed_ = true;
		} 
		else 
		{
			Module::tcl_command(args, userdata);
		}
	}
	
	void
	GenerateWireElectrode::presave()
	{
		size_t num_points = widget_.size();
		
		widget_point_.clear();
		for (size_t k = 0; k< num_points; k++)
		{
			widget_point_.push_back(new GuiPoint(get_ctx()->subVar("widget-point-"+to_string(k))));
			widget_point_[k]->set(widget_[k]->GetPosition());
		}
		
		gui_widget_points_.set(num_points);
	}
	
	void
	GenerateWireElectrode::post_read()
	{
		size_t num_points = gui_widget_points_.get();
		
		widget_point_.clear();
		std::vector<Point> points;
		
		for (size_t k = 0; k< num_points; k++)
		{
			widget_point_.push_back(new GuiPoint(get_ctx()->subVar("widget-point-"+to_string(k))));
			widget_point_[k]->reset();
			points.push_back(widget_point_[k]->get());
		}
		
		create_widgets(points);
	}
	
	void
	GenerateWireElectrode::add_point(vector<Point>& p)
	{
		
		
		size_t size=widget_.size();
		vector<Point> points(size);
		
				
		for (size_t k = 0; k < size; k++) points[k] = widget_[k]->GetPosition();
		
		p.resize(size+1);
		
		for (size_t k = 0; k < size-1; k++) p[k]=points[k];
		
		p[size]=points[size-1];
		p[size-1]=Point(points[size-2]+(points[size-1]-points[size-2])*.5);
		
		
				
	}
	
	
	bool
	GenerateWireElectrode::remove_point()
	{
		//cout<<"---removing a widget with remove button"<<endl;
				
		if (widget_.size() > 3)
		{
			widget_.pop_back();
		}
		else 
		{
			error("Must have at least 3 points.");
		}
		
		want_to_execute();
			
		return (true);
	}
	
	
}

		
		
		