/*
   For more information, please see: http://software.sci.utah.edu

   The MIT License

   Copyright (c) 2009 Scientific Computing and Imaging Institute,
   University of Utah.

   
   Permission is hereby granted, free of charge, to any person obtaining a
   copy of this software and associated documentation files (the "Software"),
   to deal in the Software without restriction, including without limitation
   the rights to use, copy, modify, merge, publish, distribute, sublicense,
   and/or sell copies of the Software, and to permit persons to whom the
   Software is furnished to do so, subject to the following conditions:

   The above copyright notice and this permission notice shall be included
   in all copies or substantial portions of the Software.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
   OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
   THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
   LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
   FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
   DEALINGS IN THE SOFTWARE.
*/



/*
 *  InterfaceWithCubit.cc:  
 *
 *  Written by:
 *   McKay Davis
 *   Scientific Computing and Imaging INstitute
 *   University of Utah
 *   May 2005
 *
 */

#include <Dataflow/Network/Module.h>
#include <Dataflow/Network/Ports/FieldPort.h>
#include <Core/Util/StringUtil.h>

#include <Core/Datatypes/SparseRowMatrix.h>
#include <Core/Datatypes/Field.h>
#include <Core/Datatypes/FieldInformation.h>

#include <Dataflow/GuiInterface/GuiVar.h>
#include <Core/ImportExport/ExecConverter.h>
#include <string>
#include <vector>
#include <map>
#include <iostream>
#include <fstream>
#include <sstream>


namespace SCIRun {

class InterfaceWithCubit : public Module
{
  public:
    InterfaceWithCubit(GuiContext* ctx);
    virtual ~InterfaceWithCubit() {}

    virtual void execute();
    
  private:
    int         field1_generation_;
    GuiString		cubitdir_;
    GuiString		ncdump_;
    string      base_;
    
    void			cleanup(string error="");
    bool			write_facet_file(string, FieldHandle &);
    bool			read_netcdf_file(string, FieldHandle &);
    bool			write_journal_file(string, string, string);

};


DECLARE_MAKER(InterfaceWithCubit)

InterfaceWithCubit::InterfaceWithCubit(GuiContext* ctx)
  : Module("InterfaceWithCubit", ctx, Filter, "NewField", "SCIRun"),
    field1_generation_(-1),
    cubitdir_(get_ctx()->subVar("cubitdir"), "."),
    ncdump_(get_ctx()->subVar("ncdump"), "ncdump")
{
}


void
InterfaceWithCubit::cleanup(string error) 
{
  unlink((base_+"exodus").c_str());
  unlink((base_+"facet").c_str());
  unlink((base_+"journal").c_str());
  unlink((base_+"netcdf").c_str());
  unlink((base_+"log").c_str());
  if (!error.empty())
    throw error;
}


bool
InterfaceWithCubit::write_facet_file(string filename, FieldHandle& handle)
{
  FILE *f = fopen(filename.c_str(), "wt");
  if (f == NULL)
  {
    error("Unable to open file"+filename+" for writing.");
    return false;
  }
  VMesh* tsm = handle->vmesh();
  
  VMesh::Node::size_type tsmns;
  VMesh::Elem::size_type tsmfs;
  VMesh::Node::iterator tsmn, tsmne;
  VMesh::Elem::iterator tsmf, tsmfe;
  VMesh::Node::array_type face_nodes;
  tsm->size(tsmns);
  tsm->size(tsmfs);
  tsm->begin(tsmn);
  tsm->end(tsmne);
  tsm->begin(tsmf);
  tsm->end(tsmfe);

  fprintf (f,"%d %d\n", static_cast<int>(tsmns), static_cast<int>(tsmfs));
  Point p;
  
  while (tsmn != tsmne) 
  {
    tsm->get_point(p, *tsmn);
    fprintf (f, "%d %f %f %f\n", static_cast<int>((*tsmn).index_), p.x(), p.y(), p.z());
    ++tsmn;
  }

  VMesh::DElem::array_type tsmea;
  VMesh::Elem::index_type neigh;
  
  tsm->synchronize(Mesh::EDGES_E | Mesh:: ELEM_NEIGHBORS_E);
		   
  while (tsmf != tsmfe) 
  {
    tsm->get_delems(tsmea, *tsmf);
    for (size_t e = 0; e < tsmea.size(); ++e) 
    {
      if (!tsm->get_neighbor(neigh, *tsmf, tsmea[e])) 
      {
        cleanup("Edge#"+to_string(tsmea[e].index_)+" is a boundary face.\n"+
        "The surface is not closed and cannot be meshed.");
      }
    }
    tsm->get_nodes(face_nodes, *tsmf);
    fprintf (f, "%d %d %d %d\n", static_cast<int>((*tsmf).index_), 
	     static_cast<int>(face_nodes[0].index_), 
       static_cast<int>(face_nodes[1].index_), 
       static_cast<int>(face_nodes[2].index_));
    ++tsmf;
  }

  fclose(f);
  remark("Wrote: "+filename);
  return true;
}
    

bool
InterfaceWithCubit::write_journal_file(string filename, string facetfilename,
				   string outputfilename)
{
  FILE *f = fopen(filename.c_str(), "wt");
  if (f == NULL)
  {
    error("Unable to open file "+filename+" for writing.");
    return false;
  }
  fprintf(f, "reset\n");
  fprintf(f, "import facet \'%s\' feature 0 make\n", facetfilename.c_str());
  fprintf(f, "volume 1 scheme tetmesh\n");
  fprintf(f, "volume 1 size 1\n");
  fprintf(f, "mesh volume 1\n");
  fprintf(f, "set large exodus file off\n");
  fprintf(f, "export mesh \'%s\' dimension 3\n", outputfilename.c_str());
  fprintf(f, "exit\n");
  fclose(f);
  remark("Wrote: "+filename);
  return true;
}

struct FillNodeFtor 
{
  Point operator()(Point &p) 
  {
    return p;
  }
};

struct FillCellFtor 
{
  FillCellFtor(vector<VMesh::index_type> &array) : array_(array) {}
  vector <VMesh::index_type> array_;
  VMesh::index_type ret_val[4];
  VMesh::index_type *operator()(VMesh::index_type n) 
  {
    for (VMesh::index_type i = 0; i < 4; ++i)
      ret_val[i] = array_[n*4+i];
    return ret_val;
  }
};

struct SimpleIter 
{
  SimpleIter(VMesh::index_type init = 0) : i(init) {}
  SimpleIter & operator++() { ++i; return *this; };
  bool operator!=(const SimpleIter &right) { return i != right.i; };
  Mesh::index_type operator-(const SimpleIter &right) { return i - right.i; };
  Mesh::index_type operator*() { return i; };
  Mesh::index_type i;
};


bool
InterfaceWithCubit::read_netcdf_file(string filename, FieldHandle& handle)
{
  ifstream infile;
  infile.open(filename.c_str());
  infile.sync();
  infile.seekg(0,ios_base::beg);
  int nnodes = -1;
  int nelems = -1;
  string instring;

  remark("Reading NetCDF file: "+filename);
  while (!infile.eof()) {
    // Eat up spurious blank lines.
    while (infile.peek() == ' ' ||
	   infile.peek() == '\t' ||
	   infile.peek() == '\n')
    {
      if (infile.bad()) return false;
      infile.get();
      if (infile.bad()) return false;
    }

    // Eat lines that start with #, they are comments.
    if (infile.peek() == '#')
    {
      if (infile.bad()) return false;
      infile.ignore(1024,'\n');
      if (infile.bad()) return false;
      continue;
    }

    if (infile.bad()) return false;

    infile >> instring;
    if (infile.bad()) return false;
    if (instring == "num_nodes") 
    {
      infile >> instring >> nnodes;
      if (infile.bad()) return false;
    } 
    else if (instring == "num_elem") 
    {
      infile >> instring >> nelems;
      if (infile.bad()) return false;
    }
    
    infile.ignore(1024,'\n');
    if (nnodes != -1 && nelems != -1) break;
  }

  if (nnodes == -1 || nelems == -1) return false;

  instring = "";
  while (infile.good() && instring != "data:") 
  {
    infile >> instring;
    if (infile.bad()) return false;
    infile.ignore(1024,'\n');
    if (infile.bad()) return false;
  }

  VMesh::index_type e, i;
  double d;
  vector <VMesh::index_type> elems;
  vector <Point> nodes;
  elems.resize(nelems*4);
  nodes.resize(nnodes);
  remark("Number of nodes in mesh: "+to_string(nnodes));
  remark("Number of elements in mesh: "+to_string(nelems));
  while (!infile.eof()) 
  {
    infile >> instring;
    if (infile.bad()) return false;
    infile.ignore(1024,'\n');    
    if (infile.bad()) return false;
    if (instring == "connect1") 
    {
      remark("Trying to read point connections.");
      for (i = 0; i < nelems*4; ++i) 
      {
        infile >> instring;
        if (infile.bad()) return false;
        if (i == nelems*4-1) instring = instring+",";
        if (!from_string(instring.substr(0, instring.size()-1),e)) 
        {
          error ("Failed to convert"+instring);
        } 
        else 
        {
          elems[i] = e-1;
        }
      }
      remark("Successfully read connecitons.");
    }

    if (instring == "coord") 
    {
      remark("Trying to read point coordinates.");
      for (e = 0; e < 3; ++e) 
      {
        for (i = 0; i < nnodes; ++i) 
        {
          infile >> instring;
          if (infile.bad()) return false;
          if (e == 2 && i == nnodes-1) instring = instring+",";
          if (!from_string(instring.substr(0,instring.size()-1),d)) 
          {
            error ("Failed to convert "+instring);
          } 
          else 
          {
            nodes[i](e) = d;
          }
        }
      }
      remark("Successfully read coordinates.");
    }
  }
  
  FieldInformation fi("TetVolMesh",-1,"double");
  handle = CreateField(fi);
  
  VMesh *tvm = handle->vmesh(); 
  
  for (size_t j=0; j<nodes.size(); j++)
  {
    tvm->add_point(nodes[j]);
  }
  
  VMesh::Node::array_type array(4);
  for (size_t j=0; j<nelems;j++)
  {
    array[0] = elems[4*j];
    array[1] = elems[4*j+1];
    array[2] = elems[4*j+2];
    array[3] = elems[4*j+3];
    tvm->add_elem(array);
  }

  return true;
}


void
InterfaceWithCubit::execute()
{
  FieldHandle field1;
  get_input_handle("Field", field1, true);

  FieldInformation fi(field1);
  
  if (!(fi.is_trisurfmesh())) 
  {
    error("Field connected to port 1 must be TriSurfField.");
    return;
  }

  // Get the PointCloudField from the second port.
  // TODO:  Why is this here?  It's not used.
  FieldHandle pcf;
  get_input_handle("PointCloudField", pcf,true);

  FieldInformation pfi(pcf);
  
  // Make sure the second input field is of type PointCloudField
  if (!(pfi.is_pointcloudmesh()))
  {
    error("Field connected to port 1 must be PointCloudField.");
    return;
  }
  
  string tmp(sci_getenv("SCIRUN_TMP_DIR")+string("/"));
  //  string rand(to_string(int(drand48()*1000.00)));
  base_ = tmp+get_id()+"."+string(sci_getenv("USER"))+".";//+rand+".";

  string facetfile = base_+"facet";
  string journalfile = base_+"journal";
  string exodusfile = base_+"exodus";
  string netcdffile = base_+"netcdf";
  string logfile = base_+"log";

  if (!write_facet_file(facetfile, field1)) 
  {
    cleanup("Cannot write facet file: "+facetfile);
  }

  if (!write_journal_file(journalfile, facetfile, exodusfile)) 
  {
    cleanup("Cannot write journal file: "+journalfile);
  }
  
  string command = 
    "export LD_LIBRARY_PATH="+cubitdir_.get()+"/libs:"+cubitdir_.get()+"/bin;"+
    cubitdir_.get()+"/bin/clarox "+journalfile+";";
  
  if (!Exec_execute_command(this, command, logfile))
  {
    cleanup("The Cubit program failed to run for some unknown reason.");
  }

  command = 
    ncdump_.get()+" -v connect1,coord "+exodusfile+" > "+netcdffile+
    "; echo Done converting "+exodusfile;

  if (!Exec_execute_command(this, command, netcdffile))
  {
    cleanup("The ncdump program failed to run for some unknown reason.");
  }

  FieldHandle output;
  if (!read_netcdf_file(netcdffile, output)) 
  {
    cleanup("Error read NetCDF file: "+netcdffile);
  }

  cleanup();
  send_output_handle("Field", output);
}

}
