/*
 For more information, please see: http://software.sci.utah.edu
 
 The MIT License
 
 Copyright (c) 2009 Scientific Computing and Imaging Institute,
 University of Utah.
 
 
 Permission is hereby granted, free of charge, to any person obtaining a
 copy of this software and associated documentation files (the "Software"),
 to deal in the Software without restriction, including without limitation
 the rights to use, copy, modify, merge, publish, distribute, sublicense,
 and/or sell copies of the Software, and to permit persons to whom the
 Software is furnished to do so, subject to the following conditions:
 
 The above copyright notice and this permission notice shall be included
 in all copies or substantial portions of the Software.
 
 THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
 THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 DEALINGS IN THE SOFTWARE.
 */

#include <Core/Datatypes/FieldInformation.h>
#include <Core/Datatypes/Field.h>
#include <Core/Datatypes/Mesh.h>
#include <math.h>
#include <Dataflow/Network/Ports/FieldPort.h>
#include <Dataflow/Network/Module.h>
#include <float.h>
#include <vector>

namespace SCIRun {
	
	class ProjectLocalMaxOnSurface : public Module {
	public:
		ProjectLocalMaxOnSurface(GuiContext*);
		virtual ~ProjectLocalMaxOnSurface() {}
		virtual void execute();
		
		GuiString gui_method_;
		
		GuiInt gui_boxsize_;
		GuiInt gui_lateraldistance_;
		GuiInt gui_projectiondistance_;
		GuiInt gui_numlayers_;  
		GuiInt gui_thresh_;
	};
	
	
	DECLARE_MAKER(ProjectLocalMaxOnSurface)
	
	ProjectLocalMaxOnSurface::ProjectLocalMaxOnSurface(GuiContext* ctx) :
	Module("ProjectLocalMaxOnSurface", ctx, Source, "ChangeFieldData", "SCIRun"),
	gui_boxsize_(ctx->subVar("boxsize"),8),
	gui_lateraldistance_(ctx->subVar("lateraldistance"),2),
	gui_projectiondistance_(ctx->subVar("projectiondistance"),3),
	gui_numlayers_(ctx->subVar("numlayers"),3),
	gui_thresh_(ctx->subVar("thresh"),100),
	gui_method_(ctx->subVar("ed-method"))
	{
	}
		
	void
	ProjectLocalMaxOnSurface::execute()
	{
		FieldHandle source, destination, result;
		
		get_input_handle("Source",source,true);
		get_input_handle("Destination",destination,true);
		
		if (inputs_changed_ || !oport_cached("Result") || gui_boxsize_.changed()
		 || gui_lateraldistance_.changed() || gui_projectiondistance_.changed()
		 || gui_numlayers_.changed() || gui_thresh_.changed() || gui_method_.changed())
		{
			update_state(Executing);
			
			FieldInformation fis(source);
			
			if (!(fis.is_latvolmesh()))
			{
				error("Source needs to be a LatVolMesh");
				return;
			}
			
			if (!(fis.is_constantdata()))
			{
				error("Source needs to have data at the elements");
				return;    
			}
			
			FieldInformation fid(destination);
			if (!(fid.is_trisurfmesh()))
			{
				error("Destination needs to be a TriSurfMesh");
				return;
			}
			// Algorithm goes here....
			
			fid.make_lineardata();
			fid.make_double();
			
			// Generate output field
			result = CreateField(fid,destination->mesh());
			
			if (result.get_rep() == 0)
			{
				error("Could not allocate result field");
				return;
			}
			
			VMesh* dmesh = destination->vmesh();
			VMesh* smesh = source->vmesh();
			
			VField* sfield = source->vfield();
			VField* rfield = result->vfield();
			
			dmesh->synchronize(Mesh::NORMALS_E);
			smesh->synchronize(Mesh::ELEM_LOCATE_E);
			
			VMesh::Node::size_type num_nodes = dmesh->num_nodes();
			
			size_type boxsize = gui_boxsize_.get();
			size_type lateraldistance = gui_lateraldistance_.get();
			size_type projectiondistance = gui_projectiondistance_.get();
			size_type numlayers = gui_numlayers_.get();
			size_type thresh = gui_thresh_.get();
			
			double val;
			double maxval;
			Point  point;
			Vector normal;
			Point  neighbor_point;
			double Node_Lateral_Distance = 0;
			double weighted_val = 0;
			double WeightedDistance = 0;
			int Layer_indicies = -1;
			double Layer_increment = double(projectiondistance)/double(numlayers);
			double Weight = 0;
			double WeightedDist = 0;
			double d = 0;
			std::vector<double> WA_Numerator;
			WA_Numerator.resize(numlayers);
			std::vector<double> WA_Denomenator;
			WA_Denomenator.resize(numlayers);
			int cnt = 0;
			
			for (VMesh::Node::index_type idx=0;idx<num_nodes;idx++)
			{
				dmesh->get_center(point,idx);
				dmesh->get_normal(normal,idx);
				index_type i,j,k;
				
				VMesh::dimension_type dims;
				
				// Get the dimensions of the grid
				smesh->get_elem_dimensions(dims);
				index_type ni = dims[0];
				index_type nj = dims[1];
				index_type nk = dims[2];
				
				VMesh::Elem::index_type elem_idx;
				VMesh::Elem::index_type neighbor_elem_idx;
				
				if(smesh->locate(elem_idx,point))
				{
					// Find the box that we are scanning through
					smesh->from_index(i,j,k,elem_idx);
					
					index_type bii = i-boxsize;
					if (bii < 0) bii = 0;
					index_type eii = i+boxsize;
					if (eii >= ni) eii = ni-1;
					index_type bjj = j-boxsize;
					if (bjj < 0) bjj = 0;
					index_type ejj = j+boxsize;
					if (ejj >= nj) ejj = nj-1;
					index_type bkk = k-boxsize;
					if (bkk < 0) bkk = 0;
					index_type ekk = k+boxsize;
					if (ekk >= nk) ekk = nk-1;
					
					// The most negative number in doubles -(DBL_Max)
					maxval = -(DBL_MAX);
					
					// For calculating weighted average
					Node_Lateral_Distance = 0;
					weighted_val = 0;
					WeightedDistance = 0;
					Layer_indicies = -1;
					for (index_type ll=0; ll<numlayers; ll++)
					{
						WA_Numerator [ll] = 0;
						WA_Denomenator [ll] = 0;
					}
					for (index_type ii = bii; ii<=eii; ii++)
					{
						for (index_type jj = bjj; jj<=ejj; jj++)
						{
							for (index_type kk = bkk; kk<=ekk; kk++)
							{
								smesh->to_index(neighbor_elem_idx,ii,jj,kk);
								smesh->get_center(neighbor_point,neighbor_elem_idx);
								Weight = 0;
								WeightedDist = 0;
								d = Dot(neighbor_point-point,normal); 
								for (k=0; k<=numlayers; k++)
								{
									if (d > k*Layer_increment && d <= (k+1)*Layer_increment)
										Layer_indicies = k;
								}
								
								if (!(Layer_indicies == -1))
								{
									
									//Node_Lateral_Distance = ((neighbor_point-point)-d*normal).length2();
									//if ( Node_Lateral_Distance < lateral_distance2)
									Node_Lateral_Distance = sqrt(((neighbor_point-point)-d*normal).length2());
									if (Node_Lateral_Distance < lateraldistance)
									{
										sfield->get_value(val,neighbor_elem_idx);
										//WeightedDist = Node_Lateral_Distance/lateral_distance2;
										WeightedDist = Node_Lateral_Distance/lateraldistance;
										Weight = (1 - (WeightedDist))*(1 - (WeightedDist))*(1 - (WeightedDist));
										//	Sums the weighted voxel intensities for calculating the weighted ave
										WA_Numerator [Layer_indicies] += (Weight*val);
										//	Sums the weights for calculating the weighted ave
										WA_Denomenator [Layer_indicies] += Weight;
									}
								}								
							}
						}
					}
					double WA = 0;
					for (index_type yy = 0; yy<=(numlayers - 1); yy++)
					{
						if (gui_method_.get() == "projection")
						{
							WA= WA_Numerator [yy] / WA_Denomenator [yy];
							if (WA > maxval) maxval = WA;
						}
						// Tranmurality Computaion
						else if (gui_method_.get() == "transmurality")
						{
							WA= WA_Numerator [yy] / WA_Denomenator [yy];
							if (yy == 0 && WA > thresh) maxval = 33;
							if (yy == 1 && WA > thresh) maxval = 66;
							if (yy == 2 && WA > thresh && maxval == 66) maxval = 100;
						 }
					}
					
					// Assign determined value to node
					if (maxval == -(DBL_MAX)) maxval = 0;
					rfield->set_value(maxval,idx);
				}
				else
				{
					//cout << "Point " << point << endl;
					//cout << "Index " << idx << endl;
					error("Surface is not located within LatVolMesh");
					//return;
				}
				
				// progress bar
				cnt++; if (cnt == 30) { cnt = 0; update_progress(idx,num_nodes); }
			}
			send_output_handle("Result",result);
		}
	}
	
} // End namespace SCIRun


