/*
   For more information, please see: http://software.sci.utah.edu

   The MIT License

   Copyright (c) 2009 Scientific Computing and Imaging Institute,
   University of Utah.

   
   Permission is hereby granted, free of charge, to any person obtaining a
   copy of this software and associated documentation files (the "Software"),
   to deal in the Software without restriction, including without limitation
   the rights to use, copy, modify, merge, publish, distribute, sublicense,
   and/or sell copies of the Software, and to permit persons to whom the
   Software is furnished to do so, subject to the following conditions:

   The above copyright notice and this permission notice shall be included
   in all copies or substantial portions of the Software.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
   OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
   THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
   LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
   FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
   DEALINGS IN THE SOFTWARE.
*/


/*
 *  BuildBEMatrix.cc: 
 *
 *  Written by:
 *   Saeed Babaeizadeh - Northeastern University
 *   Michael Callahan - Department of Computer Science - University of Utah
 *   May, 2003
 *
 */

#include <Core/Algorithms/Math/LinearSystem/SolveLinearSystem.h>

#include <Packages/BioPSE/Core/Algorithms/NumApproximation/BuildBEMatrix.h>
#include <Dataflow/Network/Module.h>
#include <Dataflow/Network/Ports/MatrixPort.h>
#include <Dataflow/Network/Ports/FieldPort.h>
#include <Core/Datatypes/DenseMatrix.h>
#include <Core/Datatypes/ColumnMatrix.h>
#include <Core/Datatypes/Field.h>
#include <Core/Datatypes/Mesh.h>

#include <Core/Util/StringUtil.h>

#include <Core/Math/MiscMath.h>
#include <math.h>

#include <algorithm>
#include <map>
#include <iostream>
#include <string>
#include <fstream>

#define epsilon 1e-12

namespace BioPSE {

using namespace SCIRun;
using namespace SCIRunAlgo;

class BuildBEMatrix : public Module, public BuildBEMatrixBase
{
  MatrixHandle       hZoi_;
  typedef LockingHandle<DenseMatrix>     DenseMatrixHandle;

private:

  bool ray_triangle_intersect(double &t,
			      const Point &p,
			      const Vector &v,
			      const Point &p0,
			      const Point &p1,
			      const Point &p2) const;
  void compute_intersections(vector<pair<double, int> > &results,
			     const VMesh* mesh,
			     const Point &p, const Vector &v,
			     int marker) const;

  int compute_parent(const vector<VMesh*> &meshes, int index);

  bool compute_nesting(vector<int> &nesting,
		       const vector<VMesh*> &meshes);

  void build_Zoi( const vector<VMesh*> &,
                                vector<int> &,
                                vector<double>&,
                                int hs,
                                int ms,
                                MatrixHandle &);

public:

  //! Constructor
  BuildBEMatrix(GuiContext *context);

  //! Destructor
  virtual ~BuildBEMatrix() {};

  virtual void execute();
};

DECLARE_MAKER(BuildBEMatrix)


unsigned int first_time_run = 1;
vector<int> field_generation_no_old, old_nesting;
vector<double> old_conductivities;

BuildBEMatrix::BuildBEMatrix(GuiContext *context):
  Module("BuildBEMatrix", context, Source, "Forward", "BioPSE")
{

}

void
BuildBEMatrix::execute()
{
  std::vector<FieldHandle> inputs;
  get_dynamic_input_handles("Surface",inputs,true);

  // Gather up the surfaces from the input ports.
  std::vector<FieldHandle> fields;
  std::vector<VMesh*> meshes;
  std::vector<double> conductivities;
  std::vector<int> field_generation_no_new;
  std::string condStr; double condVal;

  int input=-1, output=-1;

  for (size_t j=0; j<inputs.size(); j++)
  {
    FieldHandle field = inputs[j];

    VMesh* mesh = field->vmesh();
    if (!(mesh->is_trisurfmesh()))
    {
      error("Surface port '" + to_string(j) +
            "' does not contain a TriSurfField");
      return;
    }

    if (!field->get_property("Inside Conductivity", condStr))
    {
      error("The 'Inside Conductivity' of the Surface port '" + to_string(j) + "' was not set. It assumes to be zero!");
      condVal = 0;
    }
    else from_string(condStr,condVal);

    if (field->get_property("in/out", condStr))
       if (condStr == "in")
         input = j;
       else if  (condStr == "out")
         output = j;

    fields.push_back(field);
    meshes.push_back(mesh);
    conductivities.push_back(Abs(condVal));
    if(first_time_run)
    {
      field_generation_no_old.push_back(-1);
      old_conductivities.push_back(-1);
      old_nesting.push_back(-1);
    }
    field_generation_no_new.push_back(field->generation);
  }

  first_time_run = 0;

  if (input==-1 || output==-1)
  {
    error(" You must define one source as the 'input' and another one as the 'output' ");
    return;
  }

  // Compute the nesting tree for the input meshes.
  std::vector<int> nesting;
  if (!compute_nesting(nesting, meshes))
  {
    error("Unable to compute a valid nesting for this set of surfaces.");
  }
  
  if (nesting[input] == (int)nesting.size())
    conductivities[output] = 0; // for ESI problem: the outermost surface is the input
  else
    conductivities[input] = 0; // does not matter, set to 0 only for the program to work right

  // Check to see if the input fields are new
  int new_fields = 0, new_nesting = 0;
  double new_conductivities = 0;
  int no_of_fields =  nesting.size();
  for (int i=0; i < no_of_fields; i++)
  {

    new_fields += Abs( field_generation_no_new[i] - field_generation_no_old[i] );
    new_nesting += Abs( nesting[i] - old_nesting[i] );
    new_conductivities += Abs( conductivities[i] - old_conductivities[i] );
    field_generation_no_old[i] = field_generation_no_new[i];
    old_nesting[i] = nesting[i];
    old_conductivities[i] = conductivities[i];
  }

  if(new_fields>(no_of_fields+2) || new_nesting || new_conductivities )  // If the input fields are new
    build_Zoi(meshes, nesting, conductivities, input, output, hZoi_);
  else
    remark("Field inputs are old. Resending stored matrix.");

  // -- sending handles to cloned objects
  MatrixHandle hzoi(hZoi_->clone());
  send_output_handle("BEM Forward Matrix", hzoi);

  return;
}

// C++ized MollerTrumbore97 Ray Triangle intersection test.
#define EPSILON 1.0e-6
bool
BuildBEMatrix::ray_triangle_intersect(double &t,
				       const Point &point,
				       const Vector &dir,
				       const Point &p0,
				       const Point &p1,
				       const Point &p2) const
{
  // Find vectors for two edges sharing p0.
  const Vector edge1 = p1 - p0;
  const Vector edge2 = p2 - p0;

  // begin calculating determinant - also used to calculate U parameter.
  const Vector pvec = Cross(dir, edge2);

  // if determinant is near zero, ray lies in plane of triangle.
  const double det = Dot(edge1, pvec);
  if (det > -EPSILON && det < EPSILON)
  {
    return false;
  }
  const double inv_det = 1.0 / det;

  // Calculate distance from vert0 to ray origin.
  const Vector tvec = point - p0;

  // Calculate U parameter and test bounds.
  const double u = Dot(tvec, pvec) * inv_det;
  if (u < 0.0 || u > 1.0)
  {
    return false;
  }

  // Prepare to test V parameter.
  const Vector qvec = Cross(tvec, edge1);

  // Calculate V parameter and test bounds.
  const double v = Dot(dir, qvec) * inv_det;
  if (v < 0.0 || u + v > 1.0)
  {
    return false;
  }

  // Calculate t, ray intersects triangle.
  t = Dot(edge2, qvec) * inv_det;

  return true;
}


void
BuildBEMatrix::compute_intersections(std::vector<std::pair<double, int> >
              &results,
				      const VMesh* mesh,
				      const Point &p, const Vector &v,
				      int marker) const
{
  VMesh::Face::iterator itr, eitr;
  mesh->begin(itr);
  mesh->end(eitr);
  double t;
  while (itr != eitr)
  {
    VMesh::Node::array_type nodes;
    mesh->get_nodes(nodes, *itr);
    Point p0, p1, p2;
    mesh->get_center(p0, nodes[0]);
    mesh->get_center(p1, nodes[1]);
    mesh->get_center(p2, nodes[2]);
    if (ray_triangle_intersect(t, p, v, p0, p1, p2))
    {
      results.push_back(pair<double, int>(t, marker));
    }
    ++itr;
  }
}

static bool
pair_less(const std::pair<double, int> &a,
	  const std::pair<double, int> &b)
{
  return a.first < b.first;
}


int
BuildBEMatrix::compute_parent(const std::vector<VMesh*> &meshes,
                              int index)
{
  Point point;
  meshes[index]->get_center(point, VMesh::Node::index_type(0));
  Vector dir(1.0, 1.0, 1.0);
  std::vector<std::pair<double, int> > intersections;

  unsigned int i;
  for (i = 0; i < (unsigned int)meshes.size(); i++)
  {
    compute_intersections(intersections, meshes[i], point, dir, i);
  }

  std::sort(intersections.begin(), intersections.end(), pair_less);

  std::vector<int> counts(meshes.size(), 0);
  for (i = 0; i < intersections.size(); i++)
  {
    if (intersections[i].second == index)
    {
      // First odd count is parent.
      for (int j = i-1; j >= 0; j--)
      {
        if (counts[intersections[j].second] & 1)
        {
          return intersections[j].second;
        }
      }
      // No odd parent, is outside.
      return (int)meshes.size();
    }
    counts[intersections[i].second]++;
  }

  // Indeterminant, we should intersect with ourselves.
  return (int)meshes.size();
}



bool
BuildBEMatrix::compute_nesting(vector<int> &nesting,
				const vector<VMesh*> &meshes)
{
  nesting.resize(meshes.size());

  unsigned int i;
  for (i = 0; i < (unsigned int)meshes.size(); i++)
  {
    nesting[i] = compute_parent(meshes, i);
  }

  return true;
}


void BuildBEMatrix::build_Zoi(const std::vector<VMesh*> &meshes,
                        std::vector<int> &nesting,
                        std::vector<double> &conductivities,
                        int hs,
                        int ms,
                        MatrixHandle &hZoi_){

  // hs is the surface whose potentials are given and ms is the surface whose potentials are desired.

  std::vector<DenseMatrixHandle> PP;
  DenseMatrixHandle h_PP_;
  int i ,j;
  double in_cond, out_cond, op_cond;
  int no_of_fields = nesting.size();

  // PP Begin
  for (i=0; i<no_of_fields; i++)
     for (j=0; j<no_of_fields; j++)
      {
        if (i==j)
          {
            in_cond = conductivities[i];
            if (nesting[i] == no_of_fields)  // The outermost surface is different from the others.
               {out_cond = 0;
                op_cond = in_cond;}
            else
               {out_cond = conductivities[nesting[i]];
                op_cond = out_cond;}
            BuildBEMatrix::make_auto_P(meshes[i], h_PP_, in_cond, out_cond, op_cond);
          }
        else
          {
            in_cond = conductivities[j];
            if (nesting[j] == no_of_fields) out_cond = 0;
            else out_cond = conductivities[nesting[j]];

            if (nesting[i] == no_of_fields) op_cond = conductivities[i];
            else op_cond = conductivities[nesting[i]];

            BuildBEMatrix::make_cross_P(meshes[i], meshes[j], h_PP_, in_cond, out_cond, op_cond);
          }
        PP.push_back(h_PP_);
      }

  // PP End

  std::vector<DenseMatrixHandle> GG;
  DenseMatrixHandle h_GG_;

  // GG Begin
  std::vector<double>   avInn_;
  int given_potentials_surface_index = hs; // This is the surface whose potentials are given i.e. heart surface
  BuildBEMatrix::pre_calc_tri_areas(meshes[given_potentials_surface_index], avInn_);

  for (i=0; i<no_of_fields; i++)
  {
    if (i==given_potentials_surface_index)
    {
      in_cond = conductivities[i];
      if (nesting[i] == no_of_fields)  // The Outermost surface is different from the others.
         {out_cond = 0;
          op_cond = in_cond;}
      else
         {out_cond = conductivities[nesting[i]];
          op_cond = out_cond;}

      BuildBEMatrix::make_auto_G(meshes[i], h_GG_, in_cond, out_cond, op_cond, avInn_);
    }
    else
    {
      in_cond = conductivities[given_potentials_surface_index];
      if (nesting[given_potentials_surface_index] == no_of_fields) out_cond = 0;
      else out_cond = conductivities[nesting[given_potentials_surface_index]];

      if (nesting[i] == no_of_fields) op_cond = conductivities[i];
      else op_cond = conductivities[nesting[i]];

      BuildBEMatrix::make_cross_G(meshes[i], meshes[given_potentials_surface_index], h_GG_, in_cond, out_cond, op_cond, avInn_);
    }
    GG.push_back(h_GG_);
  }

  if (no_of_fields>2)
  {

    // Page 288, Pilkington Paper, IEEE Trans. on Biomedical Engineering, Vol. BME-33. No.3, March 1986
    DenseMatrixHandle hPhh_ = PP[hs*(no_of_fields+1)];
    DenseMatrixHandle hPmm_ = PP[ms*(no_of_fields+1)];
    DenseMatrixHandle hGmh_ = GG[ms];
    DenseMatrixHandle hGhh_ = GG[hs];
    DenseMatrixHandle hPmh_ = PP[ms*no_of_fields+hs];
    DenseMatrixHandle hPhm_ = PP[hs*no_of_fields+ms];

    //DenseMatrix *omatrix = 0;
    VMesh::Node::size_type nsize_hs, nsize_ms, nsize;
    meshes[given_potentials_surface_index]->size(nsize_hs);
    meshes[ms]->size(nsize_ms);

    DenseMatrixHandle omatrix;
    DenseMatrixHandle hGth_;
    int first = 1;
    for (i=0; i<no_of_fields; i++)
    {
      if (i!=hs && i!=ms)
      {
        if(first)
        {
          first = 0;
          hGth_ = GG[i];
        }
        else
        {
          omatrix=new DenseMatrix(hGth_->nrows()+GG[i]->nrows(),nsize_hs);
          Concat_rows(*omatrix.get_rep() ,*hGth_.get_rep(), *GG[i].get_rep());
          hGth_ = omatrix;
        }
      }
    }

    DenseMatrixHandle hPht_;
    first = 1;
    for (i=hs*no_of_fields; i<no_of_fields*(hs+1); i++)
    {
      if (i!=hs*(no_of_fields+1) && i!=hs*no_of_fields+ms)
      {
        if(first)
        {
          first = 0;
          hPht_ = PP[i];
        }
        else
        {
          omatrix=new DenseMatrix(nsize_hs,hPht_->ncols()+PP[i]->ncols());
          Concat_cols(*omatrix.get_rep() ,*hPht_.get_rep(), *PP[i].get_rep());
          hPht_ = omatrix;
        }
      }
    }

    DenseMatrixHandle hPmt_;
    first = 1;
    for (i=ms*no_of_fields; i<no_of_fields*(ms+1); i++)
    {
      if (i!=ms*no_of_fields+hs && i!=ms*(no_of_fields+1))
      {
        if(first)
        {
          first = 0;
          hPmt_ = PP[i];
        }
        else
        {
          omatrix=new DenseMatrix(nsize_ms,hPmt_->ncols()+PP[i]->ncols());
          Concat_cols(*omatrix.get_rep() ,*hPmt_.get_rep(), *PP[i].get_rep());
          hPmt_ = omatrix;
        }
      }
    }

    DenseMatrixHandle hPth_;
    first = 1;
    for (i=0; i<no_of_fields; i++)
    {
      if (i!=ms && i!=hs)
      {
        if(first)
        {
          first = 0;
          hPth_ = PP[i*no_of_fields+hs];
        }
        else
        {
          omatrix=new DenseMatrix(hPth_->nrows()+PP[i*no_of_fields+hs]->nrows(),nsize_hs);
          Concat_rows(*omatrix.get_rep() ,*hPth_.get_rep(), *PP[i*no_of_fields+hs].get_rep());
          hPth_ = omatrix;
        }
      }
    }

    DenseMatrixHandle hPtm_;
    first = 1;
    for (i=0; i<no_of_fields; i++)
    {  
      if (i!=ms && i!=hs)
      {
        if(first)
        {
          first = 0;
          hPtm_ = PP[i*no_of_fields+ms];
        }
        else
        {
          omatrix=new DenseMatrix(hPtm_->nrows()+PP[i*no_of_fields+ms]->nrows(),nsize_ms);
          Concat_rows(*omatrix.get_rep() ,*hPtm_.get_rep(), *PP[i*no_of_fields+ms].get_rep());
          hPtm_ = omatrix;
        }
      }
    }

    DenseMatrixHandle hPtt_, hPtt_row_;
    int firstrow = 1;

    for (i=0; i<no_of_fields; i++)
    {
      if (i!=ms && i!=hs)
      {
        meshes[i]->size(nsize);
        first = 1;
        for (j=i*no_of_fields; j<no_of_fields*(i+1); j++)
        {
          if (j!=i*no_of_fields+hs && j!=i*no_of_fields+ms)
          {
            if(first)
            {
              first = 0;
              hPtt_row_ = PP[j];
            }
            else
            {
              omatrix=new DenseMatrix(nsize,hPtt_row_->ncols()+PP[j]->ncols());
              Concat_cols(*omatrix.get_rep() ,*hPtt_row_.get_rep(), *PP[j].get_rep());
              hPtt_row_ = omatrix;
            }
          }
        }
        // One rwo of Ptt is ready
        if(firstrow)
        {
          firstrow = 0;
          hPtt_ = hPtt_row_;
        }
        else
        {
          omatrix=new DenseMatrix(hPtt_row_->nrows()+hPtt_->nrows(), hPtt_row_->ncols());
          Concat_rows(*omatrix.get_rep() ,*hPtt_.get_rep(), *hPtt_row_.get_rep());
          hPtt_ = omatrix;
        }
      }
    }
    hGhh_->invert();

    DenseMatrixHandle hG_m_h_ = new DenseMatrix(hGmh_->nrows(), hGmh_->ncols());
    Mult(*(hG_m_h_.get_rep()), *(hGmh_.get_rep()), *(hGhh_.get_rep())); // G_m_h <- Gmh*Ghh^-1

    DenseMatrixHandle hG_t_h_ = new DenseMatrix(hGth_->nrows(), hGth_->ncols());
    Mult(*(hG_t_h_.get_rep()), *(hGth_.get_rep()), *(hGhh_.get_rep())); // G_t_h <- Gth*Ghh^-1

    DenseMatrixHandle htmpMM_ = new DenseMatrix(hPmm_->nrows(), hPmm_->ncols());
    Mult(*htmpMM_.get_rep(), *hG_m_h_.get_rep(), *hPhm_.get_rep());   // tmpMM <- Gmh*Ghh^-1*Phm
    Add(1, *(hPmm_.get_rep()), -1, *(htmpMM_.get_rep()));    // Pmm <- Pmm-Gmh*Ghh^-1*Phm

    DenseMatrixHandle htmpMT_ = new DenseMatrix(hPmt_->nrows(), hPmt_->ncols());
    Mult(*htmpMT_.get_rep(), *hG_m_h_.get_rep(), *hPht_.get_rep());   // tmpMT <- Gmh*Ghh^-1*Pht
    Add(1, *(hPmt_.get_rep()), -1, *(htmpMT_.get_rep()));    // Pmt <- Pmt-Gmh*Ghh^-1*Pht

    DenseMatrixHandle htmpTT_ = new DenseMatrix(hPtt_->nrows(), hPtt_->ncols());
    Mult(*htmpTT_.get_rep(), *hG_t_h_.get_rep(), *hPht_.get_rep());   // tmpTT <- Gth*Ghh^-1*Pht
    Add(1, *(hPtt_.get_rep()), -1, *(htmpTT_.get_rep()));    // Ptt <- Ptt-Gth*Ghh^-1*Pht
    hPtt_->invert();
    htmpTT_ = 0; // Freeing the memory

    DenseMatrixHandle htmpTM_ = new DenseMatrix(hPtm_->nrows(), hPtm_->ncols());
    Mult(*htmpTM_.get_rep(), *hG_t_h_.get_rep(), *hPhm_.get_rep());   // tmpTM <- Gth*Ghh^-1*Phm
    Add(-1, *(hPtm_.get_rep()), 1, *(htmpTM_.get_rep()));    // Ptm <- -Ptm+Gth*Ghh^-1*Phm
    htmpTM_ = 0; // Freeing the memory

    Mult(*htmpMT_.get_rep(), *hPmt_.get_rep(), *hPtt_.get_rep());   // tmpMT <- (Pmt-Gmh*Ghh^-1*Pht)*(Ptt-Gth*Ghh^-1*Pht)^-1
    Mult(*htmpMM_.get_rep(), *htmpMT_.get_rep(), *hPtm_.get_rep());   // tmpMM <- (Pmt-Gmh*Ghh^-1*Pht)*(Ptt-Gth*Ghh^-1*Pht)^-1*(-Ptm+Gth*Ghh^-1*Phm)
    Add(1, *(hPmm_.get_rep()), 1, *(htmpMM_.get_rep()));
    hPmm_->invert();    // Pmm <- Part A

    DenseMatrixHandle htmpMH_ = new DenseMatrix(hPmh_->nrows(), hPmh_->ncols());
    Mult(*htmpMH_.get_rep(), *hG_m_h_.get_rep(), *hPhh_.get_rep());   // tmpMH <- Gmh*Ghh^-1*Phh
    Add(-1, *(hPmh_.get_rep()), 1, *(htmpMH_.get_rep()));    // Pmh <- -Pmh+Gmh*Ghh^-1*Phh

    DenseMatrixHandle htmpTH_ = new DenseMatrix(hPth_->nrows(), hPth_->ncols());
    Mult(*htmpTH_.get_rep(), *hG_t_h_.get_rep(), *hPhh_.get_rep());   // tmpTH <- Gth*Ghh^-1*Phh
    Add(1, *(hPth_.get_rep()), -1, *(htmpTH_.get_rep()));    // Pth <- Pth-Gth*Ghh^-1*Phh

    Mult(*htmpMH_.get_rep(), *htmpMT_.get_rep(), *hPth_.get_rep());   // tmpMH <- (Pmt-Gmh*Ghh^-1*Pht)*(Ptt-Gth*Ghh^-1*Pht)^-1*(Pth-Gth*Ghh^-1*Phh)
    htmpMT_ = 0;  // Freeing the memory

    Add(1, *(hPmh_.get_rep()), 1, *(htmpMH_.get_rep()));    // Pmh <- Part B

    DenseMatrixHandle hZmh_ = new DenseMatrix(hPmh_->nrows(), hPmh_->ncols());
    Mult(*hZmh_.get_rep(), *hPmm_.get_rep(), *hPmh_.get_rep());

    hZoi_ = hZmh_.get_rep();

  } // end   if (no_of_fields>1)
  else   //  for just two surfaces: heart and body
  {
  // deflation for just two surfaces (method used in Jeroen's Matlab code):
  double Pbbcols=PP[ms*(no_of_fields+1)]->ncols(), Phhcols=PP[hs*(no_of_fields+1)]->ncols();
  double deflate=1/(Pbbcols+Phhcols);
  //printf("deflate=%f\n",deflate);
  // deflate Pbb
  for(int i=0;i<PP[ms*(no_of_fields+1)]->nrows();i++)
  {
	for(int j=0;j<PP[ms*(no_of_fields+1)]->ncols();j++)
	{
		PP[ms*(no_of_fields+1)]->add(i,j,deflate);
	}
  }
  // deflate Phb
  for(int i=0;i<PP[hs*(no_of_fields-1)+1]->nrows();i++)
  {
	for(int j=0;j<PP[hs*(no_of_fields-1)+1]->ncols();j++)
	{
		PP[hs*(no_of_fields-1)+1]->add(i,j,deflate);
	}
  }
  // deflate Pbh
  for(int i=0;i<PP[ms*(no_of_fields-1)+1]->nrows();i++)
  {
	for(int j=0;j<PP[ms*(no_of_fields-1)+1]->ncols();j++)
	{
		PP[ms*(no_of_fields-1)+1]->add(i,j,deflate);
	}
  }
  // deflate Phh
  for(int i=0;i<PP[hs*(no_of_fields+1)]->nrows();i++)
  {
	for(int j=0;j<PP[hs*(no_of_fields+1)]->ncols();j++)
	{
		PP[hs*(no_of_fields+1)]->add(i,j,deflate);
	}
  }
  
  
  // ------------------------------------------------
  // ------------------------------------------------
  // Burak's Debugging Edits:
  
  // -- create handles for matrices that correspond to Jeroen's code
  DenseMatrixHandle Pbb(PP[ms*(no_of_fields+1)]->clone());
  DenseMatrixHandle Phb(PP[hs*(no_of_fields-1)+1]->clone());
  DenseMatrixHandle Pbh(PP[ms*(no_of_fields-1)+1]->clone());
  DenseMatrixHandle Phh(PP[hs*(no_of_fields+1)]->clone());
  
  DenseMatrixHandle Ghh(GG[hs]->clone());
  DenseMatrixHandle Gbh(GG[ms]->clone());
  
  Ghh->invert();
  
  DenseMatrixHandle Y = new DenseMatrix(Gbh->nrows(),Ghh->ncols());
  Mult(*(Y.get_rep()), *(Gbh.get_rep()), *(Ghh.get_rep()));
  DenseMatrixHandle C = new DenseMatrix(Pbb->nrows(),Pbb->ncols());
  Mult(*(C.get_rep()), *(Y.get_rep()), *(Phb.get_rep())); // C <- Y*Phb
  Sub(*(C.get_rep()),*(Pbb.get_rep()),*(C.get_rep())); // C <- -C + Pbb   =>    C = Pbb-Y*Phb
  DenseMatrixHandle D = new DenseMatrix(Pbh->nrows(),Pbh->ncols());
  Mult(*(D.get_rep()), *(Y.get_rep()), *(Phh.get_rep())); // D <- Y*Phh
  Sub(*(D.get_rep()),*(D.get_rep()),*(Pbh.get_rep())); // D <- D - Pbh   =>   D = Y*Phh-Pbh
  
  C->invert();
  DenseMatrixHandle T = new DenseMatrix(C->nrows(),D->ncols());
  Mult(*(T.get_rep()), *(C.get_rep()), *(D.get_rep())); // T = inv(C)*D
  hZoi_=T.get_rep();
  
/*  
  // -- inversion by iterative solution
  SolveLinearSystemAlgo alternInv;
  alternInv.set_option("method","bicg");
  alternInv.set_scalar("target_error",1e-30);
  alternInv.set_int("max_iterations",500);
  
  MatrixHandle GhhT(Ghh->clone()->transpose());
  MatrixHandle GbhT(Gbh->clone()->transpose());
  
  //printf("\n\n\n(%d,%d) (%d,%d)\n\n\n",int(GhhT->nrows()),int(GhhT->ncols()),int(GbhT->submatrix(0,0,GbhT->nrows()-1,0)->nrows()),int(GbhT->submatrix(0,0,GbhT->nrows()-1,0)->ncols()));
  
  MatrixHandle tempInv = new DenseMatrix(GhhT->ncols(),1);
  DenseMatrixHandle Y = new DenseMatrix(GbhT->ncols(),GhhT->ncols());
  MatrixHandle init = new DenseMatrix(GhhT->ncols(),1);
  for(int i=0;i<init->nrows();i++)
  {
	init->put(i,0,1);
  }
  
  MatrixHandle convergence;
  
  //printf("Solving for Y...\n");
  for(int i=0;i<GbhT->ncols();i++)
  {
	alternInv.run(GhhT,GbhT->submatrix(0,i,GbhT->nrows()-1,i),init,tempInv,convergence);
	for(int m=0;m<Y->ncols();m++)
		Y->put(i,m,tempInv->get(m,0)); // put the values in the rows of Y
	//printf("Y %d:\t(%d,%e)\n",i+1,convergence->get_data_size(),convergence->min());
  }
  
  DenseMatrixHandle C = new DenseMatrix(Pbb->nrows(),Pbb->ncols());
  Mult(*(C.get_rep()), *(Y.get_rep()), *(Phb.get_rep())); // C <- Y*Phb
  //Add(-1,*(C.get_rep()),1,*(Pbb.get_rep())); // C <- -C + Pbb   =>    C = Pbb-Y*Phb
  Sub(*(C.get_rep()),*(Pbb.get_rep()),*(C.get_rep()));
  DenseMatrixHandle D = new DenseMatrix(Pbh->nrows(),Pbh->ncols());
  Mult(*(D.get_rep()), *(Y.get_rep()), *(Phh.get_rep())); // D <- Y*Phh
  //Add(1,*(D.get_rep()),-1,*(Pbh.get_rep())); // D <- D - Pbh   =>   D = Y*Phh-Pbh
  Sub(*(D.get_rep()),*(D.get_rep()),*(Pbh.get_rep()));
  
  MatrixHandle tempInv2 = new DenseMatrix(C->ncols(),1);
  MatrixHandle T = new DenseMatrix(C->ncols(),D->ncols());
  MatrixHandle init2 = new DenseMatrix(C->ncols(),1);
  for(int i=0;i<init2->nrows();i++)
  {
	init2->put(i,0,1);
  }

  //printf("Solving for T...\n");  
  for(int i=0;i<D->ncols();i++)
  {
	alternInv.run(C->submatrix(0,0,C->nrows()-1,C->ncols()-1),D->submatrix(0,i,D->nrows()-1,i),init2,tempInv2,convergence);
	for(int m=0;m<T->nrows();m++)
		T->put(m,i,tempInv2->get(m,0)); // put the values in the columns of T
	//printf("T %d:\t(%d,%e)\n",i+1,convergence->get_data_size(),convergence->min());
  }

  hZoi_=T;
  */
  
//  MatrixHandle Cdump(C->clone());
//  MatrixHandle Ddump(D->clone());
//  // -- sending handles to cloned objects
//  send_output_handle("Pbb", Cdump);
//  send_output_handle("Phb", Ddump);
//  send_output_handle("Pbh", Pbh);
//  send_output_handle("Phh", Phh);
//  
//  send_output_handle("Ghh", Tdump);
//  send_output_handle("Gbh", Gbh);

  // ------------------------------------------------
  // ------------------------------------------------
//    GG[hs]->invert(); // GG[hs] = hGhh_
//
//    DenseMatrixHandle hZbh_ = new DenseMatrix(GG[ms]->nrows(), GG[ms]->ncols()); // GG[ms] = hGbh_
//    Mult(*(hZbh_.get_rep()), *(GG[ms].get_rep()), *(GG[hs].get_rep())); // Zbh <- Gbh*Ghh^-1
//    DenseMatrix tmpBB(PP[ms*(no_of_fields+1)]->nrows(), PP[ms*(no_of_fields+1)]->ncols()); // PP[ms*(no_of_fields+1)] = hPbb_
//    Mult(tmpBB, *hZbh_.get_rep(), *PP[hs*(no_of_fields-1)+1].get_rep());   // tmpBB <- Gbh*Ghh^-1*Phb   PP[hs*(no_of_fields-1)+1]=hPhb_
//    Add(1, *(PP[ms*(no_of_fields+1)].get_rep()), -1, tmpBB);    // Pbb <- part A (Pbb-Gbh*Ghh^-1*Phb)
////MatrixHandle Pbb(PP[ms*(no_of_fields+1)]->clone());
////send_output_handle("Pbb", Pbb);
//    PP[ms*(no_of_fields+1)]->invert();
//
//    DenseMatrix tmpBH(PP[ms*(no_of_fields-1)+1]->nrows(), PP[ms*(no_of_fields-1)+1]->ncols()); // PP[ms*(no_of_fields-1)+1] = hPbh_
//
//    Mult(tmpBH, *hZbh_.get_rep(), *PP[hs*(no_of_fields+1)].get_rep());  // tmpBH <- Gbh*Ghh^-1*Phh   PP[hs*(no_of_fields+1)] = hPhh_
//
//    Add(-1, *(PP[ms*(no_of_fields-1)+1].get_rep()), 1, tmpBH);   // Pbh <- part B (Gbh*Ghh^-1*Phh-Pbh)
////MatrixHandle Pbh(PP[ms*(no_of_fields-1)+1]->clone());
////send_output_handle("Pbh", Pbh);
//
//    Mult(*(hZbh_.get_rep()), *(PP[ms*(no_of_fields+1)].get_rep()), *(PP[ms*(no_of_fields-1)+1].get_rep()));
//    //hZbh_->scalar_multiply(-1.0); // <- Burak: I commented this...if you restore the other negative signs, this must also be uncommented
//    hZoi_ = hZbh_.get_rep();

  } // end else   if (no_of_fields>1)

  return;
}


} // end namespace BioPSE
