/*
   For more information, please see: http://software.sci.utah.edu

   The MIT License

   Copyright (c) 2009 Scientific Computing and Imaging Institute,
   University of Utah.

   
   Permission is hereby granted, free of charge, to any person obtaining a
   copy of this software and associated documentation files (the "Software"),
   to deal in the Software without restriction, including without limitation
   the rights to use, copy, modify, merge, publish, distribute, sublicense,
   and/or sell copies of the Software, and to permit persons to whom the
   Software is furnished to do so, subject to the following conditions:

   The above copyright notice and this permission notice shall be included
   in all copies or substantial portions of the Software.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
   OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
   THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
   LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
   FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
   DEALINGS IN THE SOFTWARE.
*/



/*
 *  ReadNrrd.cc: Read in a Nrrd
 *
 *  Written by:
 *   David Weinstein
 *   School of Computing
 *   University of Utah
 *   February 2001
 *
 */

#include <Dataflow/Network/Module.h>

#include <Core/Util/StringUtil.h>
#include <Core/ImportExport/Nrrd/NrrdIEPlugin.h>
#include <Core/Util/sci_system.h>
#include <Dataflow/GuiInterface/GuiVar.h>
#include <Dataflow/Network/Ports/NrrdPort.h>

#include <sys/stat.h>
#include <sstream>

#ifdef _WIN32
#include <process.h> // for getpid
#endif

namespace SCITeem {

using namespace SCIRun;

class ReadNrrd : public Module {
public:
  ReadNrrd(SCIRun::GuiContext* ctx);
  virtual ~ReadNrrd();
  virtual void execute();

private:
  bool read_nrrd();
  bool read_file(string filename);
  bool write_tmpfile(string filename, string *tmpfilename, string conv_command);

  GuiString       types_;
  GuiString       filetype_;
  GuiFilename     filename_;
  GuiString       from_env_;
  
  NrrdDataHandle  read_handle_;

  string          old_filename_;
  time_t          old_filemodification_;
  int             cached_label_generation_;
  char *          cached_label_;
};

} // end namespace SCITeem

using namespace SCITeem;

DECLARE_MAKER(ReadNrrd)

ReadNrrd::ReadNrrd(SCIRun::GuiContext* ctx) : 
  Module("ReadNrrd", ctx, Filter, "DataIO", "Teem"),
  types_(get_ctx()->subVar("types", false)),
  filetype_(get_ctx()->subVar("filetype")),
  filename_(get_ctx()->subVar("filename"), ""),
  from_env_(get_ctx()->subVar("from-env"),""),
  read_handle_(0),
  old_filemodification_(0),
  cached_label_generation_(0),
  cached_label_(0)
{
  NrrdIEPluginManager mgr;
  vector<string> importers;
  mgr.get_importer_list(importers);
  
  string importtypes = "{";
  importtypes += "{{Nrrd Files}    {.nhdr .nrrd .png .txt .vtk} } ";
  importtypes += "{{VFF}           {.vff} } "; // temporary
  importtypes += "{{PICT}          {.pic pict} } "; // temporary
  importtypes += "{{Vista File}    {.v} } "; // temporary
  importtypes += "{{NrrdData File} {.nd} } ";
  importtypes += "{{Nrrd File Any} {.*} } ";
/*
  for (unsigned int i = 0; i < importers.size(); i++)
  {
    NrrdIEPlugin *pl = mgr.get_plugin(importers[i]);
    if (pl->fileextension != "")
    {
      importtypes += "{{" + importers[i] + "} {" + pl->fileextension + "} } ";
    }
    else
    {
      importtypes += "{{" + importers[i] + "} {.*} } ";
    }
  }
*/
  importtypes += "}";

  types_.set(importtypes);
}

ReadNrrd::~ReadNrrd()
{
  if (cached_label_) { delete [] cached_label_; cached_label_ = 0; }
}


// Return true if handle_ was changed, otherwise return false.  This
// return value does not necessarily signal an error!
bool
ReadNrrd::read_nrrd() 
{
  filename_.reset();
  string fn(filename_.get());
  if (fn == "") 
  { 
    error("Please specify nrrd filename");
    return (false); 
  }

  // Read the status of this file so we can compare modification timestamps.
  struct stat buf;
  if (stat(fn.c_str(), &buf) == - 1) 
  {
    error(string("NrrdReader error - file not found: '")+fn+"'");
    return false;
  }

  // If we haven't read yet, or if it's a new filename, 
  //  or if the datestamp has changed -- then read...
  time_t new_filemodification = buf.st_mtime;

  if(!read_handle_.get_rep() || 
     fn != old_filename_ || 
     new_filemodification != old_filemodification_)
  {
    old_filemodification_ = new_filemodification;
    old_filename_=fn;
    read_handle_ = 0;

    int len = fn.size();
    // Filename as string
    const string filename(fn);
    const string ext(".nd");
    const string vff_ext(".vff");
    const string vff_conv_command("vff2nrrd %f %t");
    const string pic_ext(".pic");
    const string pic_ext2(".pict");
    const string pic_conv_command("PictToNrrd %f %t");
    const string vista_ext(".v");
    const string vista_conv_command("VistaToNrrd %f %t");


    // check that the last 3 chars are .nd for us to pio
    if (fn.substr(len - ext.size(), ext.size()) == ext) 
    {
      Piostream *stream = auto_istream(fn, this);
      if (!stream)
      {
        error("Error reading file '" + fn + "'.");
        return (true);
      }

      // Read the file
      Pio(*stream, read_handle_);
      if (!read_handle_.get_rep() || stream->error())
      {
        error("Error reading data from file '" + fn +"'.");
        delete stream;
        return (true);
      }
      delete stream;
    } 
    else 
    { // assume it is just a nrrd
      if (fn.substr(len - vff_ext.size(), vff_ext.size()) == vff_ext)
      {
        string tmpfilename;
        write_tmpfile(filename, &tmpfilename, vff_conv_command);
        return read_file(tmpfilename);		
      } 
      else if ((fn.substr(len - pic_ext.size(), pic_ext.size()) == pic_ext) ||
        fn.substr(len - pic_ext2.size(), pic_ext2.size()) == pic_ext2)
      {
        string tmpfilename;
        write_tmpfile(filename, &tmpfilename, pic_conv_command);
        return read_file(tmpfilename);
      } 
      else if (fn.substr(len - vista_ext.size(), vista_ext.size()) == vista_ext)
      {
        string tmpfilename;
        write_tmpfile(filename, &tmpfilename, vista_conv_command);
        return read_file(tmpfilename);	
      } 
      else 
      {
        return read_file(fn);
      }
    }
    return true;
  }
    return false;
}


bool
ReadNrrd::read_file(string fn)
{
  // Restrict TEEM access: it is not thread safe
  NrrdData::lock_teem();
  NrrdDataHandle n = new NrrdData;
  if (nrrdLoad(n->nrrd_, airStrdup(fn.c_str()), 0)) 
  {
    // Ugly error handling
    char *err = biffGetDone(NRRD);
    error("Read error on '" + fn + "': " + err);
    free(err);
    NrrdData::unlock_teem();
    return (true);
  }
  
  read_handle_ = n;
  
  NrrdData::unlock_teem();
  return (false);
}


bool
ReadNrrd::write_tmpfile(string filename, string* tmpfilename,
                          string conv_command)
{
  string::size_type loc = filename.find_last_of("/");
  const string basefilename =
    (loc==string::npos)?filename:filename.substr(loc+1);
	
  // Base filename with first extension removed.
  loc = basefilename.find_last_of(".");
  const string basenoext = basefilename.substr(0, loc);

  // Filename with first extension removed.
  loc = filename.find_last_of(".");
  const string noext = filename.substr(0, loc);

  // Temporary filename.
  //tmpfilename = "/tmp/" + basenoext + "-" +
  //  to_string((unsigned int)(getpid())) + ".nhdr";
  string tmpdir = string(sci_getenv("SCIRUN_TMP_DIR")) + "/";
  tmpfilename->assign(tmpdir + basenoext + "-" +
		      to_string((unsigned int)(getpid())) + ".nhdr");
  
  ASSERT(sci_getenv("SCIRUN_OBJDIR"));
  string command =
    string(sci_getenv("SCIRUN_OBJDIR")) + "/convert/" +
    conv_command;
  while ((loc = command.find("%f")) != string::npos)
  {
    command.replace(loc, 2, '"'+filename+'"');
  }
  while ((loc = command.find("%t")) != string::npos)
  {
    command.replace(loc, 2, '"'+*tmpfilename+'"');
  }
  const int status = sci_system(command.c_str());
  if (status)
  {
    remark("'" + command + "' failed.  Read may not work.");
  }
  return true;
}


void
ReadNrrd::execute()
{
  update_state(NeedData);

  read_nrrd();

  if (!read_handle_.get_rep()) 
  { 
    error("Please load a nrrd.");
    return; 
  }

  // A hack to make PowerApps at least work with old types of Nrrds
  Nrrd* nrrd = read_handle_->nrrd_;
  if (nrrd->spaceDim == 0)
  {
    size_t dim = nrrd->dim;
    for (size_t j=0; j<dim;j++)
    {
      if (IsNan(nrrd->axis[j].min) && IsNan(nrrd->axis[j].max)
        && !IsNan(nrrd->axis[j].spacing))
      {
        nrrd->axis[j].min = 0.0;
        nrrd->axis[j].max = (nrrd->axis[j].size-1)*nrrd->axis[j].spacing;
      }
    }
  }
  
  // Send the data downstream.
  send_output_handle("Output Data", read_handle_, true);

  update_state(Completed);
}
